from django import template
from django.template.context import RequestContext
from django.template.loader import render_to_string

from advanced_reports.views import list as advreport_list
from advanced_reports import get_report_or_404

register = template.Library()
             
@register.tag(name='advreport_detail')
def advreport_detail(parser, token):
    try:
        tag_name, advreport_slug, obj_list, internal_mode, report_header_visible = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError, "%s tag requires exactly 4 argument" % token.contents.split()[0]
    return AdvReportNode(advreport_slug, obj_list, internal_mode, report_header_visible)

class AdvReportNode(template.Node):
    def __init__(self, advreport_slug, obj_list, internal_mode, report_header_visible):
        if isinstance(advreport_slug, str):
            self.advreport_slug = advreport_slug
        else:
            self.advreport_slug = template.Variable(advreport_slug)
        self.obj_list = template.Variable(obj_list)
        self.internal_mode = internal_mode
        self.report_header_visible = report_header_visible
        self.request = template.Variable('request')
    
    def render(self, context):
        try:
            if isinstance(self.advreport_slug, str):
                slug = self.advreport_slug[1:-1]
            else:
                slug = self.advreport_slug.resolve(context)
            obj_list = self.obj_list.resolve(context)
            internal_mode = self.internal_mode.lower() == 'true'
            report_header_visible = self.report_header_visible.lower() == 'true'
            request = self.request.resolve(context)
            ids = []
            if isinstance(obj_list, list):
                ids = [o.pk for o in obj_list]
            else:
                ids = [obj_list.id,]
            return advreport_list(request, slug, ids, internal_mode, report_header_visible)
        except Exception, e:
             return e

@register.tag(name='advreport_js')
def advreport_js(parser, token):
    return AdvReportJSNode()

class AdvReportJSNode(template.Node):
    def __init__(self):
        self.request = template.Variable('request')

    def render(self, context):
        request = self.request.resolve(context)
        return render_to_string('advanced_reports/inc_js.html', {}, context_instance=RequestContext(request))

@register.tag(name='advreport_css')
def advreport_css(parser, token):
    return AdvReportCSSNode()

class AdvReportCSSNode(template.Node):
    def __init__(self):
        self.request = template.Variable('request')

    def render(self, context):
        request = self.request.resolve(context)
        return render_to_string('advanced_reports/inc_css.html', {}, context_instance=RequestContext(request))


import json
from django import forms
from django.contrib.auth.models import User
from django.http.response import HttpResponse
from django.test import TestCase
from django.test import client
from django.test.client import RequestFactory
from django.test.utils import override_settings
import six
import advanced_reports
from advanced_reports.backoffice.examples.reports import UserForm

from advanced_reports.defaults import BootstrapReport, action, ActionException
from advanced_reports.backoffice.shortcuts import action as bootstrap_action
from advanced_reports.views import api_action


class TestForm(forms.Form):
    testfield = forms.CharField(required=True)


class SimpleReport(BootstrapReport):
    models = (User,)
    decorate_views = True

    def queryset(self):
        return User.objects.all()

    item_actions = (
        action(method='test', verbose_name='Test'),
        action(method='test2', verbose_name='Test2'),
        action(method='test3', verbose_name='Test3', group='test'),
        action(method='test4', verbose_name='Test4'),
    )

    def get_decorator(self):
        return lambda x: x

    def test(self, item):
        pass

    def test2_multiple(self, items):
        pass

    def test3(self, item):
        pass

    def test4_multiple(self, items):
        return HttpResponse('OK')

    @action(verbose_name='Test5')
    def test5(self, item):
        raise ActionException('test5 action failed with an ActionException')

    @action(verbose_name='Details', regular_view=True, form=TestForm)
    def details(self, item, form):
        return HttpResponse(form.cleaned_data['testfield'])

    @bootstrap_action(verbose_name='Edit', form=UserForm, prefetch_ajax_form=True)
    def edit(self, item, form):
        form.save()

    @action(verbose_name='Remove', confirm='Are you sure?')
    def remove(self, item):
        item.delete()

    @action(verbose_name='Dialog')
    def dialog(self, item):
        return 'Hello I am a dialog for %s!' % item.first_name

    def verify_action_group(self, item, group):
        if group is None: return True
        return False


advanced_reports.register(SimpleReport)


class ReportViewsTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        User.objects.all().delete()
        cls.u = User.objects.create(username='test', password='test', email='test@example.com',
                                    first_name='Test', last_name='User')
        cls.user_checkbox = 'checkbox_0000_%d' % cls.u.pk

    def setUp(self):
        self.client = client.Client()

    def test_list(self):
        response = self.client.get('/reports/simple/')
        self.assertEqual(response.status_code, 200)

    def test_list_post(self):
        response = self.client.post('/reports/simple/', {'method': 'test'})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], 'http://testserver/reports/simple/')

    def test_list_no_method(self):
        response = self.client.post('/reports/simple/', {'method': ''})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], 'http://testserver/reports/simple/')

    def test_list_multiple_action_implicit(self):
        response = self.client.post('/reports/simple/', {'method': 'test', self.user_checkbox: 'true'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('Successfully executed action on 1 user', response.content)

    def test_list_multiple_action_explicit(self):
        response = self.client.post('/reports/simple/', {'method': 'test2', self.user_checkbox: 'true'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('Successfully executed action on 1 user', response.content)

    def test_list_multiple_action_no_selection(self):
        response = self.client.post('/reports/simple/', {'method': 'test2'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('You did not select any user.', response.content)

    def test_list_multiple_action_none_applicable(self):
        response = self.client.post('/reports/simple/', {'method': 'test3', self.user_checkbox: 'true'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('No selected user is applicable for this action.', response.content)

    def test_list_multiple_action_response(self):
        u = User.objects.latest('pk')
        response = self.client.post('/reports/simple/', {'method': 'test4', self.user_checkbox: 'true'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual('OK', response.content)

    def test_list_multiple_action_exception(self):
        response = self.client.post('/reports/simple/', {'method': 'test5', self.user_checkbox: 'true'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertIn('test5 action failed with an ActionException', response.content)

    def test_list_csv(self):
        response = self.client.get('/reports/simple/?csv', follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content,
                         'Username;Email address;First name;Last name;Staff status\n' +
                         'test;test@example.com;Test;User;False\n')

    def test_api_action_simple(self):
        response = self.client.post('/reports/api/simple/action/test/%d/' % self.u.pk)
        self.assertIn('application/json', response['Content-Type'])
        data = json.loads(response.content)
        self.assertIn('success', data)
        self.assertIn('item', data)
        self.assertNotIn('removed_item_id', data)

    @override_settings(DEBUG=True)
    def test_api_action_regular_view_with_form(self):
        post_data = {'%d-testfield' % self.u.pk: 'test input'}

        # Post valid form input which should validate
        response = self.client.post(
            '/reports/api/simple/action/details/%d/' % self.u.pk,
            post_data,
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )

        self.assertIn('application/json', response['Content-Type'])
        data = json.loads(response.content)
        self.assertIn('link_action', data)
        self.assertEqual(data['link_action']['method'], 'details')
        self.assertEqual(data['link_action']['data'], post_data)

        # Now we try form input that does not validate
        response = self.client.post(
            '/reports/api/simple/action/details/%d/' % self.u.pk,
            {},
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        data = json.loads(response.content)
        self.assertEqual(data['response_method'], 'details')
        self.assertIn('response_form', data)
        self.assertNotIn('success', data)

        # Now we use the form data to actually get the normal view (note that we do
        # a regular request here, not an AJAX one).
        response = self.client.post(
            '/reports/api/simple/action/details/%d/' % self.u.pk,
            post_data
        )
        self.assertIn('text/html', response['Content-Type'])
        self.assertContains(response, 'test input')

    def test_api_action_exception(self):
        response = self.client.post('/reports/api/simple/action/test5/%d/' % self.u.pk)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, 'test5 action failed with an ActionException')

    def test_api_action_removed_item(self):
        u2 = User.objects.create_user('test2', 'test2@example.com', 'test2')
        response = self.client.post('/reports/api/simple/action/remove/%d/' % u2.pk)
        self.assertIn('application/json', response['Content-Type'])
        data = json.loads(response.content)
        self.assertIn('success', data)
        self.assertIn('item', data)
        self.assertIsNone(data['item'])
        self.assertIn('removed_item_id', data)
        self.assertEqual(data['removed_item_id'], six.text_type(u2.pk))

    def test_api_form(self):
        response = self.client.post('/reports/api/simple/form/details/%d/' % self.u.pk)
        self.assertContains(response, '%d-testfield' % self.u.pk)
        response = self.client.post('/reports/api/simple/form/details/')
        self.assertContains(response, 'actionform-testfield')
        response = self.client.post('/reports/api/simple/form/edit/%d/' % self.u.pk)
        self.assertContains(response, '%d-first_name' % self.u.pk)
        self.assertContains(response, 'value="Test"')
        self.assertContains(response, 'value="User"')
        response = self.client.post('/reports/api/simple/form/nonexisting/%d/' % self.u.pk)
        self.assertEqual(response.status_code, 404)

    def test_string_response_dialog(self):
        request = RequestFactory().post('/reports/api/simple/action/dialog/%d/' % self.u.pk)
        response = api_action(request, 'simple', 'dialog', self.u.pk)
        self.assertEqual(response['dialog_content'], 'Hello I am a dialog for Test!')

    def test_api_list(self):
        response = self.client.post('/reports/api/simple/')
        self.assertIn('application/json', response['Content-Type'])
        data = json.loads(response.content)

        self.assertEqual(data['action_list_type'], 'links')
        self.assertEqual(data['extra'], {
            'ascending': True,
            'order_by': 'username',
            'order_field': 'username',
            'ordered_by': 'Username'
        })

        self.assertIn('field_metadata', data)
        self.assertEqual(data['field_metadata']['email'], {'full_name': 'email',
                                                           'name': 'email',
                                                           'order_by': 'email',
                                                           'sortable': True,
                                                           'style': None,
                                                           'verbose_name': 'Email address'})
        self.assertEqual(data['filter_fields'], [])
        self.assertEqual(data['filter_values'], {})

        self.assertIn('header', data)
        self.assertEqual(data['header'][0], data['field_metadata']['username'])

        self.assertIsInstance(data['item_count'], int)
        self.assertIsInstance(data['items_per_page'], int)
        self.assertIn('items', data)
        self.assertEqual(data['multiple_action_list'], [{'is_regular_view': False,
                                                         'method': 'test',
                                                         'verbose_name': 'Test'},
                                                        {'is_regular_view': False,
                                                         'method': 'test2',
                                                         'verbose_name': 'Test2'},
                                                        {'group': u'test',
                                                         'is_regular_view': False,
                                                         'method': 'test3',
                                                         'verbose_name': 'Test3'},
                                                        {'is_regular_view': False,
                                                         'method': 'test4',
                                                         'verbose_name': 'Test4'},
                                                        {'is_new_style': True,
                                                         'is_regular_view': False,
                                                         'method': 'test5',
                                                         'verbose_name': 'Test5'},
                                                        {'confirm': 'Are you sure?',
                                                         'is_new_style': True,
                                                         'is_regular_view': False,
                                                         'method': 'remove',
                                                         'verbose_name': 'Remove'},
                                                        {'is_new_style': True,
                                                         'is_regular_view': False,
                                                         'method': 'dialog',
                                                         'verbose_name': 'Dialog'}])

        self.assertFalse(data['multiple_actions'])
        self.assertTrue(data['report_header_visible'])
        self.assertEqual(data['search_fields'], ['username', 'first_name', 'last_name', 'email'])
        self.assertEqual(data['searchable_columns'], 'You can search by Username, First name, Last name, Email address')
        self.assertTrue(data['show_action_bar'])



from django.contrib import messages
from django.http.request import QueryDict
from django.http.response import HttpResponseBase
from django.shortcuts import redirect
from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.utils.translation import ugettext as _
from advanced_reports.backoffice.base import BackOfficeView
from advanced_reports import get_report_for_slug
from advanced_reports.defaults import ActionException
from advanced_reports.views import api_list, api_action, api_form


class AdvancedReportView(BackOfficeView):
    """
    A BackOffice view that renders an Advanced Report.

    Usage:
    <div view="advanced_report" params="{slug: 'report_slug', updateLocation: true|false}"></div>

    ``slug`` is the slug of your registered Advanced Report

    ``updateLocation`` is a boolean. If true, the location bar will be updated with querystring parameters
    reflecting the current filters and ordering.
    """
    slug = 'advanced_report'
    template = 'advanced_reports/backoffice/contrib/advanced-reports/advanced-report.html'

    def get(self, request):
        report_slug = request.view_params.get('slug')
        advreport = get_report_for_slug(report_slug)
        context = {'advreport': advreport}
        return render_to_string(self.template, context, context_instance=RequestContext(request))

    def fetch(self, request):
        obj_id = request.view_params.get('pk', None)
        ids = [obj_id] if obj_id else None
        return api_list(request, request.view_params.get('slug'), ids=ids)

    def form(self, request):
        method = request.action_params.get('method')
        pk = request.action_params.get('pk')
        slug = request.view_params.get('slug')
        return api_form(request, slug, method, pk and int(pk) or None)

    def action(self, request):
        method = request.action_params.get('method')
        pk = request.action_params.get('pk')
        slug = request.view_params.get('slug')
        data = request.action_params.get('data')
        if data:
            # We have to do str(data) because otherwise QueryDict is too lazy to decode...
            post = QueryDict(str(data), encoding='utf-8')
            request.POST = post
        return api_action(request, slug, method, pk and int(pk) or None)

    def action_view(self, request):
        report_slug = request.view_params.get('slug')
        method = request.view_params.get('report_method')
        pk = request.view_params.get('pk')
        return api_action(request, report_slug, method, pk and int(pk) or None)

    def multiple_action(self, request):
        report_slug = request.view_params.get('slug')
        method = request.action_params.get('report_method')
        items = request.action_params.get('items').split(',')
        global_select = request.action_params.get('global')
        advreport = get_report_for_slug(report_slug)
        data = request.action_params.get('data')
        if data:
            # We have to do str(data) because otherwise QueryDict is too lazy to decode...
            post = QueryDict(str(data), encoding='utf-8')
            request.POST = post

        if global_select:
            items, context = advreport.get_object_list(request)
        else:
            items = [advreport.get_item_for_id(pk) for pk in items]
        if hasattr(advreport, '%s_multiple' % method):
            try:
                action = advreport.find_action(method)
                if action.form:
                    form = action.form(request.POST, prefix='actionform')
                    if form.is_valid():
                        if action.is_regular_view and request.is_ajax():
                            return {'link_action': {'method': method, 'data': request.POST}}
                        response = getattr(advreport, '%s_multiple' % method)(items, form)
                    else:
                        if action.form_template:
                            response_form = render_to_string(action.form_template, {'form': form})
                        else:
                            response_form = unicode(form)
                        return {'response_form': response_form}
                else:
                    response = getattr(advreport, '%s_multiple' % method)(items)
                if response:
                    return response
                messages.success(request, _(u'Successfully executed action on all selected items.'))
            except ActionException, e:
                messages.error(request, e.msg)
            return {'succeeded': {}}
        else:
            succeeded, failed = {}, {}
            for item in items:
                try:
                    action = advreport.find_object_action(item, method)
                    if action and action.is_allowed(request):
                        result = getattr(advreport, method)(item)
                        if isinstance(result, HttpResponseBase) and result.status_code == 200:
                            messages.warning(request, _(u'This action does not support batch operations.'))
                        else:
                            succeeded[advreport.get_item_id(item)] = action.get_success_message()
                    elif not action.is_allowed(request):
                        failed[advreport.get_item_id(item)] = _(u'You are not allowed to execute this action.')
                    else:
                        failed[advreport.get_item_id(item)] = _(u'This action is not applicable to this item.')
                except ActionException, e:
                    failed[advreport.get_item_id(item)] = e.msg
            if succeeded and not failed:
                messages.success(request, _(u'Successfully executed action on all selected items.'))
            elif succeeded and failed:
                messages.warning(request, _(u'Some actions were successful, but some were also failed.'))
            else:
                messages.error(request, _(u'No action on the selected items was successful.'))
            return {'succeeded': succeeded, 'failed': failed}

    def multiple_action_view(self, request):
        report_slug = request.view_params.get('slug')
        method = request.view_params.get('report_method')
        items = request.view_params.get('items').split(',')
        global_select = request.view_params.get('global')
        advreport = get_report_for_slug(report_slug)

        if global_select == 'true':
            items = advreport.get_object_list(request)[0]
        else:
            items = [advreport.get_item_for_id(pk) for pk in items]

        items = [item for item in items if advreport.find_object_action(item, method)]
        if items:
            action = advreport.find_action(method)
            if action.form:
                form = action.form(request.GET, prefix='actionform')
                if form.is_valid():
                    return getattr(advreport, '%s_multiple' % method)(items, form)
            else:
                return getattr(advreport, '%s_multiple' % method)(items)
        messages.error(request, _(u'No items were applicable for this action.'))
        return redirect(request.META['HTTP_REFERER'])

    def auto_complete(self, request):
        partial = request.action_params.pop('partial')
        report_slug = request.view_params.get('slug')

        advreport = get_report_for_slug(report_slug)
        return advreport.auto_complete(request, partial, request.action_params)


class AdvancedReportActionView(BackOfficeView):
    slug = 'actionview'

    def get(self, request):
        report_slug = request.view_params.get('slug')
        method = request.view_params.get('method')
        pk = request.view_params.get('pk')

        advreport = get_report_for_slug(report_slug)
        item = advreport.get_item_for_id(pk)
        advreport.enrich_object(item, request=request)
        response = getattr(advreport, method)(item)
        return response.content



from __future__ import unicode_literals


try:
    from swampdragon.models import SelfPublishModel
except ImportError:
    class SelfPublishModel(object):
        pass

try:
    from swampdragon.serializers.model_serializer import ModelSerializer
except ImportError:
    class ModelSerializer(object):
        pass



from __future__ import unicode_literals
from django.conf.urls import url
from django.template.response import TemplateResponse
from django.views.generic.base import TemplateView
import advanced_reports
from advanced_reports.backoffice.base import BackOfficeBase
from advanced_reports.backoffice.contrib.views import AdvancedReportView, AdvancedReportActionView
from advanced_reports.backoffice.examples.backoffice import UserModel, UserView
from advanced_reports.backoffice.examples.reports import NoModelReport, UserReport, NewStyleReport, TodoListReport
from advanced_reports.backoffice.examples.views import SimpleView
from advreport_examples.views import ExampleIncludePythonView, ExampleIncludeTemplateView
from oemfoe_todos_app.backoffice.definitions import TodoListModel, TodoItemModel, TodoListsView


class TodosBackoffice(BackOfficeBase):
    title = 'Oemfoe Todo List Administration'
    model_template = 'advreport_examples/page-base.html'

    def define_urls(self):
        return (
            url(r'^users/$',
                self.decorate(TemplateView.as_view(template_name='advreport_examples/users.html')),
                name='users'),
            url(r'^examples/$',
                self.decorate(TemplateView.as_view(template_name='advreport_examples/examples.html')),
                name='examples'),
        )

    def page(self, request):
        return TemplateResponse(request, 'advanced_reports/backoffice/tests/page.html', {'backoffice': self})


todos_backoffice = TodosBackoffice(name='todos')


todos_backoffice.register_model(UserModel)
todos_backoffice.register_view(UserView)
todos_backoffice.register_view(SimpleView)

todos_backoffice.register_model(TodoListModel)
todos_backoffice.register_model(TodoItemModel)
todos_backoffice.register_view(TodoListsView)

todos_backoffice.register_view(AdvancedReportView)
todos_backoffice.register_view(AdvancedReportActionView)

todos_backoffice.register_view(ExampleIncludeTemplateView)
todos_backoffice.register_view(ExampleIncludePythonView)

advanced_reports.register(NoModelReport)
advanced_reports.register(UserReport)
advanced_reports.register(NewStyleReport)
advanced_reports.register(TodoListReport)



import datetime
import time
from collections import OrderedDict

from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.db.models.query import QuerySet
from django.http.response import Http404
from django.template.context import RequestContext
from django.template.defaultfilters import capfirst
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.utils.html import strip_tags, escape
from django.utils.translation import ugettext_lazy as _
from django.forms.models import fields_for_model

import six

from advanced_reports.backoffice.base import AutoSlug


class ActionType(object):
    LINKS = 'links'
    BUTTONS = 'buttons'
    INLINE_BUTTONS = 'inline_buttons'


class action(object):
    attrs_dict = None
    creation_counter = 0

    #: Required when not using the decorator. This is a string with the method name of your AdvancedReport subclass.
    #: This method must accept an item instance. If you choose to use a form, it gets an item and a bound form.
    #: If something goes wrong in your action, raise an ActionException with a message to display to the user.
    method = None

    #: Required. This is what the user will see in the frontend. If your action is simple (as in, it has no form), this
    #: will be the name of the web link. If your action uses a form, this will be a title that will be placed before it.
    #: If you don't want to display that, just use u''.
    verbose_name = None

    #: Optional. This will be used as the success message when your action has successfully executed.
    success = None

    #: Optional. When filled in, a confirmation dialog will be displayed using this attribute as the prompt.
    confirm = None

    #: Optional. When used together with a override of ``verify_action_group``, you can define on a per-item basis
    #: which action will be shown or not. This is useful when your items have a certain state only allowing for
    #: actions directly related to that state.
    group = None

    #: Optional. A Django form class that should be used together with this action. Make sure your action method
    #: has a signature like this: ``def my_action(self, item, form):``.
    #:
    #: Actions support the use of Django forms. When assigning a form class to this action, the behavior of actions
    #: will be altered in the following ways:
    #:
    #:  * A form will be presented to the user, most likely in a pop-up (if ``form_via_ajax`` is True, this is the
    #:    default when using the ``@action`` decorator version). The submit button of the form will have the
    #:    ``submit`` attribute of this action as the caption.
    #:  * When using a ModelForm, the current item instance (if available) will be used as the form's ``instance``.
    #:    keyword argument.
    #:  * Upon submission of this form, Advanced Reports will try to validate the form. If the form is not valid,
    #:    the form errors are automatically displayed to the frontend, instead of executing the underlying action.
    #:    If the form does pass the validation, the underlying action is called with the validated form as an extra
    #:    argument. If you are using a ModelForm, you can just perform a ``form.save()`` in the action method body.
    #:
    form = None

    #: Determines whether the action form should be shown as a pop-up dialog in the frontend or not.
    #: The default value is False, but if you are using the ``@action`` decorator version, the default is True.
    #: This strange default behavior is merely for backwards compatibility.
    form_via_ajax = False

    #: Determines whether the action form should be rendered at the time of the report items generation.
    #: By default this is False, but inline action forms (when ``form_via_ajax`` is False) will always be
    #: prefetched, as they are directly visible.
    prefetch_ajax_form = False

    #: If True, the link behind the action will be loaded in a pop-up. This only works if the template
    #: supports this. An alternative system for showing pop-ups is just returning the html for this pop-up
    #: in the body of the action method. This is implemented in the AngularJS version.
    link_via_ajax = False

    #: Required when using form. This is the value of your submit button when using forms.
    submit = u'Submit'

    #: Optional. If True, the submit button will be shown. The hiding of the submit button only works
    #: when the template supports this.
    show_submit = True

    #: Optional. Only used when the template supports this.
    collapse_form = False

    #: If a template path is assigned to this attribute, this template will be rendered instead of the output
    #: of {{ form }}. The template receives ``form`` and ``item`` as context variables.
    form_template = None

    #: Optional. If set, this instance will be used to pass to a ModelForm.
    #: You can also supply a callable. The kwargs are: ``instance`` and optionally ``param``.
    form_instance = None

    #: Optional. If True, when an action was executed successfully, the current row will be collapsed and the next row
    #: will be expanded, and the first field of the first form gets the input focus.
    next_on_success = False

    #: Makes the action invisible. If you want to use this action somewhere else, you can use the
    #: {% url 'advanced_reports_action' report_slug action_method item_id %} to execute the action.
    hidden = False

    #: Optional. A custom css class for this action.
    css_class = None

    #: Whether the action is shown on individual items. It could be that an action is only designed to work on
    #: a bulk items. When setting to False, you remove them from the individual items.
    individual_display = True

    #: Enable the use of actions than can execute on multiple items.
    #: If you want to implement bulk methods operating directly on a list or queryset (if applicable), you
    #: can implement an extra ``def FOO_multiple(self, items, form=None):`` where FOO is the method name of your
    #: action. If not, it will just loop through your items and execute your action on each item one after another.
    multiple_display = True

    #: If the form of the action has a file upload, set this to True. Then the enctype="multipart/form-data"
    #: will be added to the ``<form>`` tag.
    has_file_upload = False

    #: Optional. The permission required for this action.
    permission = None

    #: Whether the action method acts like a regular Django view.
    regular_view = False

    #: Whether the action is report-wide or not. Report-wide actions methods do not
    #: receive an argument containing the item. Only a form will be given if applicable.
    #: Report-wide actions are ideal for adding new items or links to another web page.
    is_report_action = False

    #: Whether the action is defined by the new style method (as a decorator) or not.
    #: The advantage is that new cool stuff that should not be backwards compatible can be enabled when this
    #: attributes is True.
    is_new_style = False

    def __init__(self, title=None, **kwargs):
        """
        Each kwarg maps to a property above. For documentation, refer to the individual property documentation strings.
        """
        action.creation_counter += 1
        self.creation_counter = action.creation_counter

        if title is not None:
            kwargs['verbose_name'] = title

        if kwargs.get('is_report_action', False):
            kwargs['individual_display'] = False
            kwargs['multiple_display'] = False

        self.attrs_dict = {}
        for k in kwargs.keys():
            setattr(self, k, kwargs[k])
            self.attrs_dict[k] = kwargs[k]

    def copy_with_instanced_form(self, advreport, prefix, instance=None, data=None):
        new_action = action(**self.attrs_dict)
        dynamic_form = getattr(advreport, 'get_%s_form' % new_action.method, None)

        if dynamic_form is not None and callable(dynamic_form):
            new_action.form = dynamic_form(instance, prefix, data=data)

        elif self.form is not None:
            new_action.form = self.form
            if issubclass(self.form, forms.ModelForm):
                new_action.form = self.form(data=data, prefix=prefix, instance=instance)
            else:
                new_action.form = self.form(data=data, prefix=prefix)

        if new_action.form is not None:
            new_action.form_template = self.form_template
            if self.form_template:
                new_action.response_form_template = mark_safe(render_to_string(self.form_template, {'form': new_action.form, 'item': instance}))

        if instance:
            context = {'item': instance}
            context.update(instance.__dict__)
            if new_action.confirm: new_action.confirm = new_action.confirm % Resolver(context)
            if new_action.success: new_action.success = new_action.success % Resolver(context)
            if new_action.verbose_name: new_action.verbose_name = new_action.verbose_name % Resolver(context)

        return new_action

    def get_success_message(self):
        return self.success or _(u'Successfully executed "%s"') % self.verbose_name

    def get_form_instance(self, instance, *args, **kwargs):
        if not instance:
            return None
        if self.form and not issubclass(self.form, forms.ModelForm):
            return None

        if 'param' in kwargs and not kwargs['param']:
            kwargs.pop('param')

        if self.form_instance:
            return self.form_instance(instance, *args, **kwargs) \
                if callable(self.form_instance) \
                else self.form_instance
        return instance

    def get_bound_form(self, request, instance, prefix):
        if self.form is None:
            return None
        kwargs = {}
        instance = self.get_form_instance(instance)
        if instance:
            kwargs['instance'] = instance
        if request.method == 'GET':
            return self.form(request.GET, prefix=prefix, **kwargs)
        return self.form(request.POST, request.FILES, prefix=prefix, **kwargs)

    def render_form(self, request, instance, form):
        if self.form_template:
            return render_to_string(self.form_template,
                                    {'form': form, 'item': instance},
                                    context_instance=RequestContext(request))
        return six.text_type(form)

    @property
    def is_regular_view(self):
        return self.regular_view or self.method.endswith('_view')

    def is_allowed(self, request):
        return not self.permission or request.user.has_perm(self.permission)

    def __call__(self, function=None):
        """
        Allows this action to be used as a decorator. Example::

            class MyReport(AdvancedReport):
                model = User

                @action('Edit user', form=UserForm)
                def edit(self, user, form):
                    form.save()
        """

        # This dark magic is needed for Django's dirty template code.
        if function is None:
            return self

        # Now that we are using a new style of calling this action, we can safely use
        # a new set of defaults :-)
        if self.form:
            if not 'form_via_ajax' in self.attrs_dict:
                self.form_via_ajax = True
                self.attrs_dict['form_via_ajax'] = True
        self.is_new_style = True
        self.attrs_dict['is_new_style'] = True

        self.method = function.__name__
        self.attrs_dict['method'] = self.method

        @six.wraps(function)
        def decorator(*args, **kwargs):
            return function(*args, **kwargs)
        decorator.action = self
        return decorator


class ActionException:
    def __init__(self, msg=None, form=None):
        if form is not None:
            msg = u''
            for k in form.errors.keys():
                for e in form.errors[k]:
                    msg += u' ' + e
            self.msg = msg
        else:
            self.msg = u'%s' % msg


class AdvancedReport(object):
    #: Required. A unique url-friendly name for your Advanced Report
    slug = AutoSlug(remove_suffix='Report')

    #: The request. A new ``AdvancedReport`` instance is created on each request. This request will be saved to this
    #: property, so you can access it anywhere you need it.
    request = None

    #: Optional when ``model`` or ``models`` is specified. A list of fields that must be displayed on the report.
    fields = None

    #: A list of sortable fields.
    #: You can sort on multiple fields at the same time using the comma (,) separator.
    #: The __ lookup syntax is also supported.
    sortable_fields = None

    #: A list of fields that are searchable. When specified or inferred from a given model
    search_fields = None

    #: A list of fields that can be filtered.
    filter_fields = ()

    #: Optional. A dictionary with fields that will be shown as tabs. 1st dict field has to be the field to filter on.
    #: 2nd dict field is a dict with options:
    #:     - ``types``: types to filter on.
    #:     - ``images`` (optional): `default` will be used for all other fields, use same name as defined in `types` to
    #:        provide a `type` specific image.
    tabbed_filter_fields = ()

    #: Optional. A mapping of filter fields to their list of values.
    filter_values = {}

    #: Deprecated. A tuple of available actions for your report. Please use the actions as a decorator instead.
    item_actions = ()

    #: Required if no model specified. A name for an individual item. For example ``_('user')``
    verbose_name = None

    #: Required if no model specified. A plural name for individual items. For example ``_('users')``
    verbose_name_plural = None

    #: Required if no model specified. The human-friendly title of your report.
    title = None

    #: Optional. Some text to explain the purpose of the report and some extra info.
    help_text = None

    #: Optional, but strongly advised to use it. A model class. An ``AdvancedReport`` can infer a lot of stuff
    #: from your model, including but not limited to the field verbose names, sortable fields, ...
    #: This is a shortcut to the ``models`` attribute.
    model = None

    #: Optional, and not needed if ``model`` is specified. A tuple of model classes. See the documentation
    #: for the ``model`` attribute.
    models = None

    #: The maximum number of items shown on one page. When you have more than this number
    #: of items, your report will be paginated. By default this is 20.
    items_per_page = 20


    #: Optional. A tuple of link tuples. You can define some top level links for your report.
    #: A link tuple must contain two items, the name of the link and the href location.
    #: Optionally you can show some more information if you add a third item to the link tuple with
    #: a short explanation. This will be shown next or below the button, or a tooltip, as the templates can vary.
    #: Example::
    #:
    #:     links = (
    #:         (u'Print this page', 'javascript:printPage();', u'Click here to print this page'),
    #:         (u'Refresh', '.'),
    #:     )
    links = ()

    #: The default global template for rendering a report. This is only used when using the server-side
    #: Django view-based version.
    template = 'advanced_reports/default.html'

    #: The default item template. This template will be used to render an individual item.
    #: When an action performs an operation on a report item, this template will be used to re-render that
    #: particular item.
    item_template = 'advanced_reports/item_row.html'

    #: This template is used when you want your view-based report to be shown in the middle of an existing page.
    internal_template = ''

    #: Required when using get_decorator. True indicates that Advanced Reports should use the implementation
    #: of the ``get_decorator`` method.
    decorate_views = False

    #: Optional. Puts checkboxes before each item and puts a combobox with group actions on your report.
    #: Only simple (as in, not using a form) actions can be performed on multiple items.
    #:
    #: You can implement your own special action method for multiple items. See the documenation for
    #: FOO_multiple below.
    multiple_actions = False

    #: Optional. Use this if you have a custom urlname for your report. Advanced Reports then knows how to
    #: reach your report.
    urlname = None

    #: Optional. If a date field name is assigned to this, you can filter on date ranges.
    date_range = None

    #: Optional. True allows animation, but that does not work with table rows.
    animation = False

    #: Optional. The text displayed when no items are shown on your report. By default this is:
    #: There are no <verbose_name_plural> to display.
    empty_text = None

    #: Controls the visibility of the table header.
    header_visible = True

    #: Controls the visibility of the report header
    report_header_visible = True

    #: Shows the separator between actions.
    show_actions_separator = True

    #: How to display the actions.
    action_list_type = ActionType.LINKS

    #: Show the actions of an item only when the user hovers over the item.
    #: This only works when your template or report rendering implementation knows about this attribute.
    show_actions_only_on_hover = True

    #: Determines if the advanced report should display a only single items. Used when you want to display
    #: to report in an other template
    internal_mode = False

    def __init__(self, *args, **kwargs):
        self.model_admin = None

        # Expand some shortcuts
        if not self.models and self.model:
            self.models = (self.model,)

        # Add defaults from the model meta
        if self.models:
            model = self.models[0]
            if not self.verbose_name:
                self.verbose_name = model._meta.verbose_name
            if not self.verbose_name_plural:
                self.verbose_name_plural = model._meta.verbose_name_plural
            if not self.title:
                self.title = capfirst(self.verbose_name_plural)

            # Add defaults from the model admin
            model_admin = admin.site._registry.get(model)

            if model_admin:
                self.model_admin = model_admin

                if self.fields is None and model_admin.list_display:
                    self.fields = model_admin.list_display

                if self.search_fields is None and model_admin.search_fields:
                    self.search_fields = model_admin.search_fields

                if self.sortable_fields is None and model_admin.list_display:
                    self.sortable_fields = model_admin.list_display

        # Some sane defaults
        if not self.fields:
            self.fields = ()
        if not self.search_fields:
            self.search_fields = ()
        if not self.sortable_fields:
            self.sortable_fields = ()

        # Expand item_actions with the decorated versions
        item_actions = list(self.item_actions)
        for method_name in dir(self):
            if not hasattr(AdvancedReport, method_name):
                method = getattr(self, method_name)

                if callable(method) and hasattr(method, 'action'):
                    item_actions.append(method.action)

        item_actions.sort(key=lambda a: a.creation_counter)
        self.item_actions = item_actions

    def queryset(self):
        """
        The primary source of the data you will want to manage.
        """
        if self.models:
            return self.models[0].objects.all()
        return None

    def get_FOO_verbose_name(self):
        '''
        Implement this function to specify the column header for the field FOO.
        '''
        return None

    def get_FOO_html(self):
        '''
        Implement this function to specify the HTML representation for the field FOO.
        '''
        return None

    def get_FOO_decorator(self):
        '''
        Implement this function to specify a decorator function for the field FOO.
        Return a function that accepts one parameter and returns the decorated html.
        '''
        return lambda s: s

    def filter_FOO(self, qs, value):
        """
        Implement this function to specify a filter for the field FOO. This is
        useful for fields that are not a part of a model, but aggregated.

        :return: a filtered queryset
        """
        return qs

    def verify_action_group(self, item, group):
        '''
        Implement this function to verify if the given group currently applies to the given item.

        For example, if the item is a paid purchase and the group is "paid", we must return True.
        To turn of filtering of actions by groups, just don't specify the group when creating an
        action and just leave this function alone.
        '''
        return True

    def set_request(self, request):
        '''
        Set the request for this report.
        '''
        self.request = request

    #
    # The following two functions work both in tandem for naming and finding items.
    # It is recommended to override get_item_for_id as the default implementation may be a little
    # bit inefficient.
    #

    def get_item_id(self, item):
        '''
        Advanced Reports expects each item to have a unique ID. By default this is the primary key
        of a model instance, but this can be anything you want, as long it is a unicode string.
        '''
        return unicode(item.pk)

    def get_item_for_id(self, item_id):
        '''
        Advanced Reports also expects each item to be found by its unique ID. By default it does
        a lookup of the primary key of a model.

        Returns None if the item does not exist.
        '''
        try:
            return self._queryset(request=None).get(pk=item_id)
        except ObjectDoesNotExist, e:
            return None

    def get_decorator(self):
        '''
        To be used in tandem with decorate_views. Set it to True when you want to implement this function.
        Here you can return a decorator. That decorator will be used to decorate the Advanced Report views.
        This way you can easily add some permission restrictions to your reports.
        '''
        return None

    def FOO(self, item, form=None):
        '''
        The implementation of the FOO action method.
        '''

    def FOO_view(self, item, form=None):
        '''
        The implementation of the FOO action method that can return a HttpResponse
        '''

    def FOO_multiple(self, items):
        '''
        The implementation of the FOO action method with multiple items.
        This may return a HttpResponse object or just None.
        '''
        return None

    def get_FOO_form(self, item, prefix, data=None):
        """
        Instead of specifying the ``form`` attribute for an action, you can also construct your
        action form instance dynamically by implementing this method in your report.
        :param item: the item for which to construct the action form
        :param prefix: the prefix that the form should use
        :param data: the data that should be passed to the form for validation
        :return: a form instance
        """
        return None

    def get_item_class(self, item):
        '''
        Implement this to get some extra CSS classes for the item. Separate them with spaces.
        '''
        return u''

    def get_FOO_class(self, item):
        '''
        Implement this to get some extra CSS classes for the field FOO. Separate them with spaces.
        '''
        return u''

    def get_extra_information(self, item):
        '''
        Implement this to get some extra information about an item.
        This will be shown right below a data row and above the actions.

        Ajax loading of information

        You can specify some "lazy divs" that are only loaded when an action row expands.
        You must define a hidden item action with a method name that ends with '_view'.

        Example::

            def get_extra_information(self, item):
                return u'<div class="lazy" data-method="credit_view"></div>'

            item_actions = (
                action(method='credit_view', verbose_name=u'', group='mygroup', hidden=True),
            )

            def credit_view(self, item):
                balance = get_credit_balance(item)
                return render_to_response('credit_template.html', {'balance': balance})

        '''
        return u''

    def get_FOO_style(self):
        '''
        Implement this to apply some CSS to the FOO td table cell.
        For example, you can define a fixed width here.
        '''
        return u''

    def auto_complete(self, request, partial, params):
        """
        Implement this to support auto completion of certain fields.

        :param request: the HTTPRequest asking for the completion
        :param partial: the partial string that must be completed
        :param params: optional parameters
        :return: a list of possible completions. Keep this list short, all data returned will be displayed.
        """
        return []

    def enrich_list(self, items):
        '''
        Implement this to attach some extra information to each item of the given items list.

        Use self.assign_attr(item, attr_name, value) to do that, it automatically detects dicts.

        Example usage: adding state information to each item to prevent multiple queries to the State model.
        '''
        pass

    def get_item_count(self):
        '''
        Implement this if you don't use Django model instances.
        Returns the number of items in the report.
        '''
        return self._queryset(request=None).count()

    def get_template(self):
        '''
        Get the template that needs to be rendered
        '''
        if self.internal_mode:
            if self.internal_template:
                return self.internal_template
            else:
                return self.template.replace('.html', '_internal.html')
        else:
            return self.template

    def _extra_context(self, request=None):
        context = {}
        context.update(self.extra_context_request(request))
        context.update(self.extra_context())
        return context

    def extra_context(self):
        '''
        (deprecated) Implement this to define some extra context for your template.
        '''
        return {}

    def extra_context_request(self, request=None):
        '''
        Implement this to define some extra context for your template,
        based on the request.
        '''
        filter_form = self.get_filter_form()
        if filter_form is not None:
            filter_form = filter_form(request.GET or None)
        return {
            'filters_form': filter_form,
            'tabbed_filters_links': self.get_tabbed_filter_links(),
        }

    def _create_choicefield(self, choices, add_empty=False):
        if add_empty and len(choices) and choices[0][0] != '':
            choices = [('', '---')] + list(choices)
        return forms.ChoiceField(choices=choices)

    def get_filter_form(self):
        """
        Ugly way to generate generic form for filters- but I can't see better idea, how to do this
        """
        if not self.models:
            return None
        all_model_fields = {}
        for model in self.models:
            all_model_fields.update(fields_for_model(model))
        report = self
        class DynamicForm(forms.Form):
            def __init__(self, *args, **kwargs):
                super(DynamicForm, self).__init__(*args, **kwargs)
                for filter_field in report.filter_fields:
                    field_fn = getattr(self, 'get_%s_filter' % filter_field, None)
                    if field_fn is not None:
                        self.fields[filter_field] = field_fn()
                    elif filter_field in report.filter_values:
                        self.fields[filter_field] = report._create_choicefield(report.filter_values[filter_field], True)
                    elif filter_field in all_model_fields:
                        self.fields[filter_field] = report._create_choicefield(all_model_fields[filter_field].choices, True)
                    else:
                        raise ValueError("We can't get choices for %s filter" % filter_field)
        return DynamicForm

    def get_tabbed_filter_links(self):
        """
        Example::

            configured in report:

                tabbed_filter_fields = (
                    {
                        'card': {
                            'images': {
                                'default': 'default.png',
                                '2FF': '2ff-image.img',
                            },
                            'types': [
                                '2FF', '2FF/3FF', '2/3/4FF', '4FF'
                            ]
                        }
                    })

            will result in (keys are slugified, so 2FF/3FF will be 2ff3ff):

                [
                    'card': [
                        {'2FF': '2ff-image.png'},
                        {'2FF/3FF', 'default.png'},
                    ],
                ]

        :return: a dict for links with optional images.
        """
        report = self
        result = OrderedDict()
        for filter_field in report.tabbed_filter_fields:
            options = report.tabbed_filter_fields[filter_field]
            if not options or options == {} or 'types' not in options:
                raise Exception('Cannot construct tabbed filter fields!')

            values = OrderedDict()
            for field_type in options['types']:
                if 'images' in options:
                    if field_type in options['images']:
                        values[field_type] = options['images'][field_type]
                    elif 'default' in options['images'].keys():
                        values[field_type] = options['images']['default']
                    else:
                        values[field_type] = None
                else:
                    values[field_type] = None
            result[filter_field] = values.items()
        return result.iteritems()

    def get_filters_from_request(self, request):
        """
        Return a dict used as argument to Model.objects.filter() function.
        But we need to be sure that we can filter on these fields so we check if there is no filter_* function defined
        in advanced report definition
        """
        result = {}
        for filter_field in self.filter_fields:
            if filter_field in request.REQUEST and request.REQUEST[filter_field].strip() and \
                    not hasattr(self, 'filter_%s' % filter_field):
                result[filter_field] = request.REQUEST[filter_field]
        return result

    def get_filtered_items(self, queryset, params, request=None):
        filter_query = None
        date_range_query = None
        fake_fields = []
        uses_model = None

        # Extract parameters
        q = params['q'].lower() if 'q' in params else None
        exact = 'exact' in params
        filters_from_request = self.get_filters_from_request(request)

        if hasattr(queryset, 'filter') and filters_from_request:
            queryset = queryset.filter(**filters_from_request)

        if params.get('from', '') or params.get('to', ''):
            date_range_query = Q()

            if params.get('from', ''):
                from_date_struct = time.strptime(params['from'], '%Y-%m-%d')
                from_date = datetime.datetime(year=from_date_struct.tm_year,
                                              month=from_date_struct.tm_mon,
                                              day=from_date_struct.tm_mday)
            if params.get('to', ''):
                to_date_struct = time.strptime(params['to'], '%Y-%m-%d')
                to_date = datetime.datetime(year=to_date_struct.tm_year,
                                            month=to_date_struct.tm_mon,
                                            day=to_date_struct.tm_mday)
                # Date range has no hour so we add 1 day to the to_date so that we get the results of that day as well
                # eg: if we selected from: 2011-01-17 and to: 2011-01-18, then the actual date range will be:
                # between 2011-01-17 00:00 and 2011-01-19 00:00
                to_date += datetime.timedelta(days=1)

            uses_model = False

            field = self.get_model_field(self.date_range.split('__')[0])

            if field is None:
                fake_fields.append(self.date_range)
            else:
                uses_model = True

        if uses_model:
            if params.get('from', '') and params.get('to', ''):
                date_range_query = Q(**{'%s__range' % self.date_range: (from_date, to_date)})

                queryset = queryset.filter(date_range_query)
            elif params.get('from', ''):
                queryset = queryset.filter(Q(**{'%s__gte' % self.date_range: from_date}))
            elif params.get('to', ''):
                queryset = queryset.filter(Q(**{'%s__lt' % self.date_range: to_date}))

        if q:
            if uses_model is None:
                uses_model = False

            parts = q.split()
            filter_query = Q()
            for part in parts:
                part_query = Q()
                for search_field in self.search_fields:
                    field = self.get_model_field(search_field.split('__')[0])
                    if field is None:
                        fake_fields.append(search_field)
                    else:
                        uses_model = True
                        if exact:
                            part_query = part_query | Q(**{'%s__iexact' % search_field: part})
                        else:
                            part_query = part_query | Q(**{'%s__icontains' % search_field: part})
                filter_query = filter_query & part_query


        fake_found = []
        if len(fake_fields) > 0:
            self.enrich_list(queryset)
            for fake_field in fake_fields:
                for o in queryset:
                    test_string = strip_tags(self.get_item_html(fake_field, o)).lower().replace(u'&nbsp;', u' ')
                    if q == test_string if exact else q in test_string:
                        fake_found.append(int(o.pk) if uses_model else o)

        if uses_model or uses_model is None:
            # uses_model is None when none of the search parameters were found
            if fake_found:
                if filter_query:
                    filter_query = filter_query | Q(pk__in=fake_found)
                else:
                    filter_query = Q(pk__in=fake_found)

            if filter_query:
                return EnrichedQueryset(queryset.filter(filter_query), self, request=request)
            else:
                # When no filter parameter is found then we don't apply the filter_query
                return EnrichedQueryset(queryset, self, request=request)
        else:
            return EnrichedQueryset(fake_found, self, request=request)

    def _queryset(self, request):
        if hasattr(self, 'queryset_request'):
            qs = self.queryset_request(request)
            if request:
                def convert_value(k, v):
                    if k[-4:] == '__in':
                        return v.split(',')
                    if v.lower() == u'true':
                        return True
                    elif v.lower() == u'false':
                        return False
                    else:
                        return v

                if self.models:
                    fieldnames = [f.name for f in self.models[0]._meta.fields]
                    lookup = dict((k, convert_value(k, v)) for k, v in request.GET.items() if k.split('__')[0] in fieldnames)
                    if lookup:
                        qs = qs.filter(**lookup)

                    for k, v in request.GET.items():
                        filter_fn = getattr(self, 'filter_%s' % k, None)
                        if filter_fn:
                            qs = filter_fn(qs, convert_value(k, v))

            return qs
        else:
            return self.queryset()

    def get_sorted_queryset(self, by_field, request=None):
        if by_field not in ('__str__', '__unicode__'):
            field_name = by_field.split('__')[0].split(',')[0]
            field_name = field_name[1:] if field_name[0] == '-' else field_name
        else:
            field_name = 'pk'
        if self.get_model_field(field_name) is None:
            return self._queryset(request)
        return self._queryset(request).order_by(*by_field.split(','))

    def get_enriched_items(self, queryset):
        return EnrichedQueryset(queryset, self)

    def get_object_list(self, request, ids=None):
        default_order_by = ''.join(self.sortable_fields[:1])
        order_by = request.GET.get('order', default_order_by)
        context = {}
        if order_by:
            order_field = order_by.split('__')[0].split(',')[0].strip('-')
            ascending = order_by[:1] != '-'
            context.update({'order_field': order_field,
                            'ascending': ascending,
                            'order_by': order_by.strip('-'),
                            'ordered_by': self.get_ordered_by(order_by)})
            queryset = self.get_sorted_queryset(order_by, request=request)
        else:
            queryset = self._queryset(request)

        # Filter
        if ids is not None:
            queryset = queryset.filter(pk__in=ids)
        object_list = self.get_filtered_items(queryset, request.GET, request=request)

        return object_list, context

    def get_ordered_by(self, by_field):
        if by_field == '':
            return u''
        field_names = (part.strip('-') for part in by_field.split(','))
        verbose_names = (self.get_field_metadata(field_name)['verbose_name'] for field_name in field_names)
        return u', '.join(verbose_names)

    def get_action_callable(self, method):
#        if isinstance(method, basestring):
#            return getattr(self, method, lambda i, f=None: False)
#        else:
#            return method
        return getattr(self, method, lambda i, f=None: False)

    def handle_multiple_actions(self, method, selected_object_ids, request=None):
        '''
        Deprecated. Don't use anymore!
        '''
        action = self.find_action(method)
        objects = [o for o in (self.get_item_for_id(item_id) for item_id in selected_object_ids) if o]
        self.enrich_list(objects)
        for o in objects:
            self.enrich_object(o, list=False, request=request)
        objects = [object for object in objects if self.verify_action_group(object, action.group)]
        handler = getattr(self, '%s_multiple' % method, None)
        if handler:
            if len(objects) == 0:
                return None, 0
            return handler(objects), len(objects)

        count = 0
        self.enrich_list(objects)
        for object in objects:
            self.enrich_object(object, list=False, request=request)
            if self.find_object_action(object, method) is not None:
                self.get_action_callable(method)(object)
                count += 1

        return None, count

    @property
    def column_headers(self):
        return [self.get_field_metadata(field_name) for field_name in self.fields]

    @property
    def searchable_columns(self):
        if not self.search_fields:
            return None
        field_names = (s.rsplit('__', 1)[-1] for s in self.search_fields)
        field_names = u', '.join(self.get_field_metadata(field_name)['verbose_name'] for field_name in field_names)
        return _(u'You can search by %(fields)s') % {'fields': field_names}

    def get_column_values(self, item):
        for field_name in self.fields:
            yield {'html': self.get_item_html(field_name, item),
                   'class': getattr(self, 'get_%s_class' % field_name, lambda i: u'')(item),
                   'style': getattr(self, 'get_%s_style' % field_name, lambda: None)()}

    def get_model_field(self, field_name):
        if self.models is None:
            return None

        for model in self.models:
            try:
                return model._meta.get_field_by_name(field_name)[0]
            except:
                pass
        return None

    def get_field_metadata_dict(self):
        all_fields = list(self.fields) + list(self.filter_fields)
        return dict((field, self.get_field_metadata(field)) for field in all_fields)

    def get_field_metadata(self, field_name):
        verbose_name = getattr(self, 'get_%s_verbose_name' % field_name, lambda: None)()
        if verbose_name is None:
            model_field = self.get_model_field(field_name)
            if model_field is not None:
                verbose_name = model_field.verbose_name

        if verbose_name is None:
            verbose_name = capfirst(field_name.replace(u'_', u' '))

        sortable = False
        order_by = ''
        for sf in self.sortable_fields:
            fn = sf.split('__')[0].split(',')[0].strip('-')
            if fn == field_name.split('__')[0]:
                sortable = True
                order_by = sf.strip('-')

        return {'name': field_name.split('__')[0],
                'full_name': field_name,
                'verbose_name': capfirst(verbose_name),
                'sortable': sortable,
                'order_by': order_by,
                'style': getattr(self, 'get_%s_style' % field_name, lambda: None)()}

    def lookup_item_value(self, field_name, item):
        if '__' in field_name:
            field_name, remainder = field_name.split('__', 1)
            obj = getattr(item, field_name, None)
            return self.lookup_item_value(remainder, obj)
        else:
            html = getattr(item, 'get_%s_display' % field_name, lambda: None)()
            if html is None:
                html = getattr(item, field_name, None)
        return html

    def get_html_for_value(self, value):
        """
        Override this method to generate HTML for specific values.
        """
        return escape(six.text_type(value))

    def get_item_html(self, field_name, item):
        html = getattr(self, 'get_%s_html' % field_name, lambda i: None)(item)
        if html is None:
            html = self.get_html_for_value(self.lookup_item_value(field_name, item))

        decorator = getattr(self, 'get_%s_decorator' % field_name, lambda i: None)(item)
        if decorator is not None:
            html = decorator(html)

        return mark_safe(html)

    def objects(self, request=None):
        return EnrichedQueryset(self._queryset(request), self)

    def enrich_object(self, o, list=True, request=None):
        '''
        This method adds extra metadata to an item.
        When the list argument is True, enrich_list will be called on the item.
        When calling this method multiple times, it is inefficient to call enrich_list,
        as it can be run once on the whole list, so in this case set list to False.
        If supplied, the request will be attached to the item so that you can use this
        in your actions.
        '''
        if not o:
            return

        if list:
            self.enrich_list([o])

        self.assign_attr(o, 'advreport_request', request)
        self.assign_attr(o, 'advreport_column_values', [v for v in self.get_column_values(o)])
        self.assign_attr(o, 'advreport_actions', self.get_object_actions(o, request=request))
        self.assign_attr(o, 'advreport_object_id', self.get_item_id(o))
        self.assign_attr(o, 'advreport_class', self.get_item_class(o))
        self.assign_attr(o, 'advreport_extra_information', self.get_extra_information(o) % Resolver({'item': o}))

    def enrich_generic_relation(self, items, our_model, foreign_model, attr_name, fallback):
        '''
        This is a utility method that can be used to prefetch backwards generic foreign key relations.
        Use this if you have a lot of lookups of this kind when generating the report.

        Parameters:
            - items: a iterable of items where we want to attach the backwards relation object.
            - our_model: the type of items where the generic foreign key points to.
            - foreign_model: the type of items that have a generic foreign relation to our_model.
            - attr_name: specifies name of the attribute that will contain an instance of foreign_model.
            - fallback: a function that will be called when the backwards relation was not found.
              It must accept a our_model instance and must return a foreign_model instance.
        '''
        ct = ContentType.objects.get_for_model(our_model)
        pks = [item.pk for item in items]
        foreigns = foreign_model.objects.filter(content_type=ct, object_id__in=pks).select_related('content_type')
        foreign_mapping = {}
        for foreign in foreigns:
            foreign_mapping[int(foreign.object_id)] = foreign
        for item in items:
            setattr(item, attr_name, foreign_mapping.get(int(item.pk), None) or fallback(item))

    def enrich_backward_relation(self, items, foreign_model, field_name, related_name,
                                 our_index=None, our_foreign_index=None,
                                 select_related=None, many=True):
        '''
        This is a utility method that can be used to pretech backward relations.
        For instance, you can add subscription lists to each user using only one line of code.

        Parameters:
            - items: a iterable of items where we want to attach the backwards relation object.
            - foreign_model: the type of items that have a foreign relation to our_model.
            - field_name: the field name of foreign_model that points to our_model.
            - related_name: specifies name of the attribute that will contain (a list of)
              instances of foreign_model.
            - our_index: a callable that gets the id of an instance of our_model.
            - our_foreign_index: a callable that gets the id of our instance from foreign_model.
            - select_related: performs a select_related on the query on foreign_model.
            - many: if set to False, a single item will be assigned instead of a list. When
              the list is empty, None will be assigned.
        '''

        our_index = our_index or (lambda i: i.pk)
        oi = lambda i: int(our_index(i)) if our_index(i) is not None else None
        our_foreign_index = our_foreign_index or (lambda i: getattr(i, '%s_id' % field_name))
        ofi = lambda i: int(our_foreign_index(i)) if our_foreign_index(i) is not None else None

        pks = [oi(item) for item in items]
        foreigns = foreign_model.objects.filter(**{'%s__in' % field_name: pks})
        if select_related:
            foreigns = foreigns.select_related(*select_related)

        foreign_mapping = {}

        for foreign in foreigns:
            our_id = ofi(foreign)
            if our_id in foreign_mapping:
                foreign_mapping[our_id].append(foreign)
            else:
                foreign_mapping[our_id] = [foreign]

        if many:
            for item in items:
                setattr(item, related_name, foreign_mapping.get(oi(item), ()))
        else:
            for item in items:
                setattr(item, related_name, foreign_mapping.get(oi(item), (None,))[0])

    def get_object_actions(self, object, request=None):
        actions = []

        for a in self.item_actions:
            if self.verify_action_group(object, a.group) and \
                    (not request or a.is_allowed(request)):
                instance = a.form_instance(object) if a.form_instance else object
                if not a.form_via_ajax or a.prefetch_ajax_form:
                    new_action = a.copy_with_instanced_form(self, prefix=self.get_item_id(object), instance=instance)
                else:
                    # Put off fetching the instanced Form until the actual Ajax
                    # call for performance.
                    new_action = a
                if not new_action.hidden and new_action.individual_display:
                    actions.append(new_action)

        return actions

    def find_object_action(self, object, method, request=None):
        found_action = None
        for a in self.item_actions:
            if object is None or self.verify_action_group(object, a.group):
                if a.method == method:
                    found_action = a
                    break

        if request is not None:
            if found_action is None:
                raise Http404(_(u'Unsupported action method "%s".' % method))
            if not found_action.is_allowed(request):
                raise Http404(_(u'You\'re not allowed to execute "%s".' % method))

        return found_action

    def find_action(self, method):
        for a in self.item_actions:
            if a.method == method:
                return a
        return None

    @property
    def date_range_verbose_name(self):
        if not self.date_range:
            return u''
        return self.get_field_metadata(self.date_range)['verbose_name']

    def get_empty_text(self):
        if self.empty_text:
            return self.empty_text
        return _(u'There are no %(items)s to display.') % {'items': self.verbose_name_plural}

    def assign_attr(self, object, attr_name, value):
        '''
        Assigns a value to an attribute. When the given object is a dict, the value will be assigned
        as a key-value pair.
        '''
        if isinstance(object, dict):
            object[attr_name] = value
        else:
            setattr(object, attr_name, value)

    def urlize(self, urlname, kwargs):
        return lambda h: u'<a href="%(l)s">%(h)s</a>' % {'l': reverse(urlname, kwargs=kwargs), 'h': h}


class EnrichedQueryset(object):
    def __init__(self, queryset, advreport, request=None):
        self.queryset = queryset
        if isinstance(self.queryset, QuerySet):
            self.queryset.query.add_ordering('pk')
        self.advreport = advreport
        self.request = request

    def __getitem__(self, k):
        if isinstance(k, slice):
            return self._enrich_list(list(self.queryset[k.start:k.stop]))
        else:
            return self._enrich(self.queryset[k])

    def __iter__(self):
        return self.queryset.__iter__()

    def __len__(self):
        if isinstance(self.queryset, QuerySet):
            return self.queryset.count()
        return len(self.queryset)

    def iterator(self):
        if isinstance(self.queryset, QuerySet):
            it = self.queryset.iterator()
        else:
            it = self.queryset
        for i in it:
            self._enrich(i)
            yield i

    def _enrich_list(self, l):
        # We run enrich_list on all items in one pass.
        self.advreport.enrich_list(l)

        for o in l:
            # We pass list=False to prevent running enrich_list from enrich_object.
            self.advreport.enrich_object(o, list=False, request=self.request)

        return l

    def _enrich(self, o):
        self.advreport.enrich_object(o, self.request)
        return o


class Resolver(object):
    def __init__(self, context):
        self.context = context

    def __getitem__(self, k):
        from django.template import Variable
        return Variable(k).resolve(self.context)


class BootstrapReport(AdvancedReport):
    template = 'advanced_reports/bootstrap/report.html'
    item_template = 'advanced_reports/bootstrap/item.html'




from django.contrib import admin
from advanced_reports.backoffice.models import SearchIndex

class SearchIndexAdmin(admin.ModelAdmin):
    list_display = ('backoffice_instance', 'model_slug', 'model_id', 'to_index')
    list_filter = ('backoffice_instance', 'model_slug')
    search_fields = ('model_id',)

admin.site.register(SearchIndex, SearchIndexAdmin)




from django.test import TestCase

from advanced_reports.defaults import AdvancedReport


class AdvancedReportTest(TestCase):
    def setUp(self):
        self.report = AdvancedReport()

    def test_tabbed_filter_fields(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                    'default': 'img/item.png',
                    'item2': 'img/item2.png'
                },
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            }
        }

        self.assertEqual(['card'], [k for k, v in self.report.get_tabbed_filter_links()])
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', 'img/item.png'),
                ('item2', 'img/item2.png'),
                ('item3', 'img/item.png'),
                ('item4', 'img/item.png'),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))

    def test_tabbed_filter_fields_default_only(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                    'default': 'img/item2.png'
                },
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            }
        }

        self.assertEqual(['card'], [k for k, v in self.report.get_tabbed_filter_links()])
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', 'img/item2.png'),
                ('item2', 'img/item2.png'),
                ('item3', 'img/item2.png'),
                ('item4', 'img/item2.png'),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))

    def test_tabbed_filter_fields_without_default_image(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                    'item2': 'img/item2.png'
                },
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            }
        }

        self.assertEqual(['card'], [k for k, v in self.report.get_tabbed_filter_links()])
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', None),
                ('item2', 'img/item2.png'),
                ('item3', None),
                ('item4', None),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))

    def test_tabbed_filter_fields_without_image_values(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                },
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            }
        }

        self.assertEqual(['card'], [k for k, v in self.report.get_tabbed_filter_links()])
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', None),
                ('item2', None),
                ('item3', None),
                ('item4', None),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))

    def test_tabbed_filter_fields_without_images(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            }
        }

        self.assertEqual(['card'], [k for k, v in self.report.get_tabbed_filter_links()])
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', None),
                ('item2', None),
                ('item3', None),
                ('item4', None),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))

    def test_tabbed_filter_fields_without_types(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                },
            }
        }
        with self.assertRaises(Exception):
            self.report.get_tabbed_filter_links()

    def test_tabbed_filter_fields_multiple(self):
        self.report.tabbed_filter_fields = {
            'card': {
                'images': {
                    'default': 'img/item.png',
                    'item2': 'img/item2.png'
                },
                'types': [
                    'item1', 'item2', 'item3', 'item4'
                ]
            },
            'gender': {
                'images': {
                    'male': 'img/male.png',
                    'female': 'img/female.png'
                },
                'types': [
                    'male', 'female'
                ]
            }
        }

        self.assertEqual(['card', 'gender'], sorted([k for k, v in self.report.get_tabbed_filter_links()]))
        dict_iteritems = [v for k, v in self.report.get_tabbed_filter_links()]
        self.assertEqual(
            [
                ('item1', 'img/item.png'),
                ('item2', 'img/item2.png'),
                ('item3', 'img/item.png'),
                ('item4', 'img/item.png'),
            ], sorted([(k, v) for k, v in dict_iteritems[1]]))
        self.assertEqual(
            [
                ('female', 'img/female.png'),
                ('male', 'img/male.png'),
            ], sorted([(k, v) for k, v in dict_iteritems[0]]))



# -*- coding: utf-8 -*-
from django import forms
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.http import HttpResponse, Http404
from django.shortcuts import render_to_response, redirect
from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.utils.html import strip_entities, strip_tags
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

import six

from .backoffice.api_utils import JSONResponse
from .decorators import conditional_delegation, report_view
from .defaults import ActionException, Resolver
from .utils import paginate

def _get_redirect(advreport, next=None, querystring=None):
    if next:
        return redirect(next)
    if advreport.urlname:
        return redirect(reverse(advreport.urlname))
    suffix = u''
    if querystring:
        suffix = u'?%s' % querystring
    return redirect(reverse('advanced_reports_list', kwargs={'slug': advreport.slug}) + suffix)


@conditional_delegation(lambda request: 'delegate' in request.GET)
@report_view
def list(request, advreport, ids=None, internal_mode=False, report_header_visible=True):
    advreport.internal_mode = internal_mode
    advreport.report_header_visible = report_header_visible

    context = {}

    # Handle POST
    if request.method == 'POST':
        sorted_keys = [k for k in request.POST.keys()]
        sorted_keys.sort()
        selected_object_ids = [k.split('_')[2] for k in sorted_keys if 'checkbox_' in k and request.POST[k] == 'true']
        method = request.POST['method']

        if not method:
            messages.warning(request, _(u'You did not select any action.'))
            if not advreport.internal_mode:
                return _get_redirect(advreport)

        if len(selected_object_ids) == 0:
            messages.warning(request, _(u'You did not select any %(object)s.') % {'object': advreport.verbose_name})
            if not advreport.internal_mode:
                return _get_redirect(advreport)

        try:
            response, count = advreport.handle_multiple_actions(method, selected_object_ids, request)
            if response:
                return response

            if count > 0:
                messages.success(request, _(u'Successfully executed action on %(count)d %(objects)s')
                                                % {'count': count,
                                                   'objects': advreport.verbose_name_plural if count != 1 else advreport.verbose_name})
            else:
                messages.error(request, _(u'No selected %(object)s is applicable for this action.') % {'object': advreport.verbose_name})
            if not advreport.internal_mode:
                return _get_redirect(advreport, querystring=request.META['QUERY_STRING'])
        except ActionException, e:
            context.update({'error': e.msg})


    object_list, extra_context = advreport.get_object_list(request, ids=ids)
    context.update(extra_context)

    # CSV?
    if 'csv' in request.GET:
        try:
            from djprogress import with_progress # pragma: no cover
        except ImportError: # pragma: no cover
            with_progress = lambda it, **kw: it # pragma: no cover
        # Avoid microsoft SYLK problem http://support.microsoft.com/kb/215591
        _mark_safe = lambda s: s if unicode(s) != u'ID' else u'"%s"' % s
        object_count = len(object_list)
        #csv = StringIO()
        header = u'%s\n' % u';'.join(_mark_safe(c['verbose_name']) for c in advreport.column_headers)
        lines = (u'%s\n' % u';'.join((c['html'] for c in o.advreport_column_values)) \
                 for o in with_progress(object_list.iterator() \
                                            if hasattr(object_list, 'iterator') \
                                            else object_list[:],
                                        name='CSV export of %s' % advreport.slug,
                                        count=object_count))
        lines = (line.replace(u'&nbsp;', u' ') for line in lines)
        lines = (line.replace(u'&euro;', u'€') for line in lines)
        lines = (line.replace(u'<br/>', u' ') for line in lines)
        lines = (strip_entities(line) for line in lines)
        lines = (strip_tags(line).encode('utf-8') for line in lines)
        #csv.write(header)
        #csv.writelines(lines)
        response = HttpResponse('', 'text/csv')
        response['Content-Disposition'] = 'attachment; filename="%s.csv"' % advreport.slug
        response.write(header)
        for line in lines:
            response.write(line)
        return response

    # Paginate
    paginated = paginate(request, object_list, advreport.items_per_page)

    # Extra context?
    context.update(advreport._extra_context(request))

    # Render
    context.update({'advreport': advreport,
                    'paginated': paginated,
                    'object_list': object_list})

    func = render_to_string if advreport.internal_mode else render_to_response
    return func(advreport.get_template(), context, context_instance=RequestContext(request))


@report_view
def action(request, advreport, method, object_id, param=None):
    next = request.GET.get('next', None)

    object = advreport.get_item_for_id(object_id)
    advreport.enrich_object(object, request=request)

    a = advreport.find_action(method)

    if request.method == 'POST':
        if a.form is not None:
            if issubclass(a.form, forms.ModelForm):
                form = a.form(request.POST, request.FILES, instance=a.get_form_instance(object), prefix=object_id)
            else:
                form = a.form(request.POST, request.FILES, prefix=object_id)

            if form.is_valid():
                r = advreport.get_action_callable(a.method)(object, form)
                # TODO: give feedback

                return r or _get_redirect(advreport, next)
    else:
        if param:
            r = advreport.get_action_callable(a.method)(object, param)
        else:
            r = advreport.get_action_callable(a.method)(object)
        # TODO: give feedback

        return r or _get_redirect(advreport, next)


@report_view
def ajax(request, advreport, method, object_id, param=None):
    object = advreport.get_item_for_id(object_id)
    advreport.enrich_object(object, request=request)
    a = advreport.find_object_action(object, method)
    if a is None:
        return HttpResponse(_(u'Unsupported action method "%s".' % method), status=404)

    context = {'advreport': advreport}

    try:
        if request.method == 'POST' and a.form is not None:
            if issubclass(a.form, forms.ModelForm):
                form = a.form(request.POST, request.FILES, instance=a.get_form_instance(object), prefix=object_id)
            else:
                form = a.form(request.POST, request.FILES, prefix=object_id)

            if form.is_valid():
                advreport.get_action_callable(a.method)(object, form)
                object = advreport.get_item_for_id(object_id)
                context.update({'success': a.get_success_message()})
            else:
                context.update({'response_method': method, 'response_form': form})
                if a.form_template:
                    context.update({'response_form_template': mark_safe(render_to_string(a.form_template, {'form': form}))})

            advreport.enrich_object(object, request=request)
            context.update({'object': object})
            return render_to_response(advreport.item_template, context, context_instance=RequestContext(request))

        elif a.form is None:
            if param:
                advreport.get_action_callable(a.method)(object, param)
            else:
                advreport.get_action_callable(a.method)(object)
            object = advreport.get_item_for_id(object_id)
            advreport.enrich_object(object, request=request)
            context = {'object': object, 'advreport': advreport, 'success': a.get_success_message()}
            context.update({'response_method': method, 'response_form': a.form})
            if a.form_template:
                context.update({'response_form_template': mark_safe(render_to_string(a.form_template, {'form': a.form}))})

            return render_to_response(advreport.item_template, context, context_instance=RequestContext(request))

    except ActionException, e:
        return HttpResponse(e.msg, status=404)

    # a.form is not None but not a POST request
    return HttpResponse(_(u'Unsupported request method.'), status=404)


@report_view
def count(request, advreport):
    return HttpResponse(unicode(advreport.get_item_count()))


@report_view
def ajax_form(request, advreport, method, object_id, param=None):
    object = advreport.get_item_for_id(object_id)
    advreport.enrich_object(object, request=request)
    a = advreport.find_object_action(object, method)
    if a is None:
        # No appropriate action found (maybe it was filtered out?)
        raise Http404

    context = {'advreport': advreport}

    if request.method == 'POST':
        a = a.copy_with_instanced_form(advreport,
                                       prefix=object_id,
                                       instance=a.get_form_instance(object, param=param),
                                       data=request.POST)
        form = a.form
        if form.is_valid():
            r = advreport.get_action_callable(a.method)(object, form)
            object = advreport.get_item_for_id(object_id)
            advreport.enrich_object(object, request=request)
            context.update({'success': a.get_success_message(), 'object': object, 'action': a})
            response = render_to_string(advreport.item_template, context, context_instance=RequestContext(request))
            return r or JSONResponse({
                'status': 'SUCCESS',
                'content': response
            })
        else:
            context.update({'response_method': method, 'response_form': form})
            if a.form_template:
                context.update({'response_form_template': mark_safe(render_to_string(a.form_template, {'form': form, 'item': object}))})

        context.update({'object': object, 'action': a})
        return render_to_response('advanced_reports/ajax_form.html', context, context_instance=RequestContext(request))

    else:
        a = a.copy_with_instanced_form(advreport, prefix=object_id, instance=a.get_form_instance(advreport.get_item_for_id(object_id), param=param))
        context = {'object': object, 'advreport': advreport, 'success': a.get_success_message(), 'action': a}

        context.update({'response_method': method, 'response_form': a.form})
        if a.form_template:
            context.update({'response_form_template': mark_safe(render_to_string(a.form_template, {'form': a.form, 'item': object}))})

        return render_to_response(
            'advanced_reports/ajax_form.html',
            context,
            context_instance=RequestContext(request)
        )


@report_view
def api_form(request, advreport, method, object_id=None):
    object = object_id and advreport.get_item_for_id(object_id)
    if object:
        advreport.enrich_object(object, request=request)
        a = advreport.find_object_action(object, method)
    else:
        a = advreport.find_action(method)
    if a is None or not a.form:
        # No appropriate action found (maybe it was filtered out?)
        raise Http404
    instance = object_id and a.get_form_instance(advreport.get_item_for_id(object_id))
    prefix = object_id or 'actionform'
    a = a.copy_with_instanced_form(advreport, prefix=prefix, instance=instance)
    return HttpResponse(a.render_form(request, instance, a.form))


def _action_dict(request, o, action):
    d = dict(action.attrs_dict)
    if action.form:
        if not action.prefetch_ajax_form and action.form_via_ajax:
            d['form'] = True
        else:
            d['form'] = action.render_form(request, o, action.form)
    if action.confirm and o:
        context = {'item': o}
        context.update(getattr(o, '__dict__', {}))
        d['confirm'] = o and action.confirm % Resolver(context) or None
    d['is_regular_view'] = action.is_regular_view
    return d


def _item_values(request, o, advreport):
    return {
        'values': o.advreport_column_values,
        'extra_information': o.advreport_extra_information.replace('data-method="',
                                                                   'ng-bind-html-unsafe="lazydiv__%s__' % advreport.get_item_id(o)),
        'actions': [_action_dict(request, o, a) for a in o.advreport_actions],
        'item_id': advreport.get_item_id(o)
    }


def _is_allowed_multiple_action(request, advreport, action):
    form_allowed = not action.form or hasattr(advreport, '%s_multiple' % action.method)
    return not action.hidden and form_allowed and action.multiple_display and action.is_allowed(request)


@report_view
def api_list(request, advreport, ids=None):
    object_list, extra_context = advreport.get_object_list(request, ids=ids)

    paginated = paginate(request, object_list, advreport.items_per_page)

    report = {
        'header': advreport.column_headers,
        'extra': extra_context,
        'items': [_item_values(request, o, advreport) for o in paginated.object_list[:]],
        'items_per_page': advreport.items_per_page,
        'item_count': len(object_list),
        'searchable_columns': advreport.searchable_columns,
        'show_action_bar': bool(advreport.search_fields or advreport.filter_fields),
        'search_fields': advreport.search_fields,
        'filter_fields': advreport.filter_fields,
        'filter_values': advreport.filter_values,
        'action_list_type': advreport.action_list_type,
        'field_metadata': advreport.get_field_metadata_dict(),
        'report_header_visible': advreport.report_header_visible,
        'multiple_actions': advreport.multiple_actions,
        'multiple_action_list': [
            _action_dict(request, None, a) \
            for a in advreport.item_actions \
            if _is_allowed_multiple_action(request, advreport, a)
        ],
        'report_action_list': [_action_dict(request, None, a) for a in advreport.item_actions if a.is_report_action]
    }
    return JSONResponse(report)


def _prepare_string_response(response, a):
    # Back to default normal-width dialog boxes.
    if isinstance(response, six.string_types):
        if a.is_new_style:
            return {
                'dialog_content': response,
                'dialog_style': {'width': '600px'}
            }
    return response


@report_view
def api_action(request, advreport, method, object_id=None):
    # Retrieve the object on which this action will operate. This can be None.
    obj = object_id and advreport.get_item_for_id(object_id)
    advreport.enrich_object(obj, request=request)
    a = advreport.find_object_action(obj, method, request)
    action_args = []

    # Unless the action is a report-wide action, always add the obj argument.
    if not a.is_report_action:
        action_args.append(obj)

    # Start with an empty reply. We will fill this in according to the different situations.
    reply = {}

    # Build a bound form if this action has a form.
    prefix = object_id or 'actionform'
    form = a.get_bound_form(request, obj, prefix)

    # Check if the action is a regular view. If we are posting a form to it in an AJAX-call,
    # and if the form is valid, we instruct the frontend to call this API using a regular
    # GET or POST.
    if a.is_regular_view and request.is_ajax() and form and form.is_valid():
        reply.update({'link_action': {'method': method, 'data': request.POST}})
    else:
        # Check if we have a form. If so, we will add it to the arguments.
        if form:
            action_args.append(form)

        # If we are not using a form, just continue. If we do have a form, make sure it is valid...
        if not form or form.is_valid():
            # The actual action method call
            try:
                response = advreport.get_action_callable(a.method)(*action_args)
            except ActionException, e:
                return HttpResponse(e.msg, status=404)

            # If we have a response, this means we can stop the normal flow and return the response
            # to the frontend, who hopefully knows what to do with it.
            if response:
                return _prepare_string_response(response, a)

            # Add the success marker to the reply
            reply.update({'success': a.get_success_message()})

            # Retrieve the object again to get the latest version. It can also be None,
            # perhaps because the action method has deleted the object.
            obj = object_id and advreport.get_item_for_id(object_id)
        else:
            # The form was not valid, so we will present the errors to the frontend.
            # Note that we don't supply a success marker.
            reply.update({'response_method': method, 'response_form': a.render_form(request, obj, form)})

    # Attach the updated object (which could also be deleted, who knows) to the reply.
    if obj:
        advreport.enrich_object(obj, request=request)
        reply.update({'item': _item_values(request, obj, advreport)})
    elif object_id:
        reply.update({'item': None, 'removed_item_id': object_id})

    # Done!
    return JSONResponse(reply)




from decimal import Decimal
from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as make_proxy
from django.conf import settings

import json


_proxy_type = type(make_proxy('ignore this'))
def _json_object_encoder(obj):
    if isinstance(obj, _proxy_type):
        return u'%s' % obj
    elif isinstance(obj, Decimal):
        return unicode(obj)
    elif hasattr(obj, 'isoformat'):
        return obj.isoformat()
    elif hasattr(obj, 'splitlines'):
        return unicode(obj)
    else:
        return None


def to_json(obj):
    if settings.DEBUG:
        return json.dumps(obj, default=_json_object_encoder, indent=2)
    return json.dumps(obj, default=_json_object_encoder)


def JSONResponse(obj):
    return HttpResponse(to_json(obj), content_type='application/json;charset=UTF-8')


class ViewRequestParameters(object):
    def __init__(self, request):
        self.GET = request.GET
        self.POST = request.POST
        try:
            self.body = request.body
            self.json_dict = json.loads(self.body) if 'application/json' in request.META['CONTENT_TYPE'] else {}
        except:
            self.json_dict = {}

        self.fallbacks = (self.GET, self.POST, self.json_dict)
        self.list_fallbacks = (self.GET, self.POST)

    def get(self, item, default=None):
        for fallback in self.fallbacks:
            obj = fallback.get(item, None)
            if obj is not None:
                return obj
        return default

    def getlist(self, item, default=None):
        for fallback in self.list_fallbacks:
            l = fallback.getlist(item)
            if l != []:
                return l
        if default is None:
            return []
        return default

    def __repr__(self):
        return 'ViewRequestParameters(%r, %r, %r)' % (self.GET, self.POST, self.json_dict)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'SearchIndex'
        db.create_table(u'backoffice_searchindex', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('backoffice_instance', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('model_slug', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('model_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('to_index', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('search_index', self.gf('djorm_pgfulltext.fields.VectorField')(default='', null=True, db_index=True)),
        ))
        db.send_create_signal(u'backoffice', ['SearchIndex'])


    def backwards(self, orm):
        # Deleting model 'SearchIndex'
        db.delete_table(u'backoffice_searchindex')


    models = {
        u'backoffice.searchindex': {
            'Meta': {'object_name': 'SearchIndex'},
            'backoffice_instance': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'model_slug': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'search_index': ('djorm_pgfulltext.fields.VectorField', [], {'default': "''", 'null': 'True', 'db_index': 'True'}),
            'to_index': ('django.db.models.fields.TextField', [], {'blank': 'True'})
        }
    }

    complete_apps = ['backoffice']


# -*- coding: utf-8 -*
from django.contrib.auth.models import User
from django.test import TestCase

from advanced_reports.defaults import AdvancedReport, EnrichedQueryset


class AdvancedReportTest(TestCase):
    def setUp(self):
        User.objects.create_user("test2", "test2@example.com", "foobar")
        self.report = AdvancedReport()
        self.report.models = (User,)

    def test_sorting(self):
        self.assertQuerysetEqual(User.objects.order_by('pk'), self.report.get_sorted_queryset('__unicode__'), transform=lambda x:x)
        self.assertQuerysetEqual(User.objects.order_by('pk'), self.report.get_sorted_queryset('__str__'), transform=lambda x:x)
        self.assertQuerysetEqual(User.objects.order_by('first_name'), self.report.get_sorted_queryset('first_name'), transform=lambda x:x)
        self.assertQuerysetEqual(User.objects.all(), self.report.get_sorted_queryset('field_that_does_not_exist'), transform=lambda x:x, ordered=False)
        self.assertQuerysetEqual(User.objects.all(), self.report.get_sorted_queryset('field_that__does_not_exist'), transform=lambda x:x, ordered=False)

    def test_enriched_queryset_order(self):
        eqs = EnrichedQueryset(User.objects.all(), self.report)
        self.assertListEqual(['pk'], eqs.queryset.query.order_by)
        eqs = EnrichedQueryset(User.objects.all().order_by('first_name'), self.report)
        self.assertListEqual(['first_name', 'pk'], eqs.queryset.query.order_by)
        eqs = EnrichedQueryset(['a','b','c'], self.report)
        self.assertListEqual(['a','b','c'], eqs.queryset)



from collections import defaultdict
from django.conf.urls import patterns, url
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.db.models.signals import post_save, post_delete
from django.http import Http404, HttpResponse
from django.http.response import HttpResponseForbidden
from django.shortcuts import render_to_response, get_object_or_404, redirect, render
from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.views.decorators.cache import never_cache

from .api_utils import JSONResponse, ViewRequestParameters
from .conf import DB_IS_POSTGRES
from .models import SearchIndex
from .search import convert_to_raw_tsquery
from .decorators import staff_member_required

import random


def random_token():
    return u'<!-- random: %d -->' % random.randint(0, 10000000000)


def check_permission(request, permission):
    """
    Checks whether the current request has access to the given permission.
    If the ``permission`` argument is ``None``, and the request is of a logged in user, we return ``True``.

    :param request HttpRequest: The current request
    :param permission str: The permission as a string of '<app name>.<permission codename>'.
    :return bool: True if the request has access to the given permission.
    """
    if not request.user.is_authenticated():
        return False
    if permission is None:
        return True
    return request.user.has_perm(permission)


class AutoSlug(object):
    def __init__(self, remove_suffix=None):
        self.remove_suffix = remove_suffix

    def __get__(self, instance, owner):
        slug = owner.__name__
        if self.remove_suffix and slug.endswith(self.remove_suffix):
            slug = slug[:-len(self.remove_suffix)]
        return slug.lower()


class AutoTemplate(AutoSlug):
    def __init__(self, directory, remove_suffix=None):
        super(AutoTemplate, self).__init__(remove_suffix=remove_suffix)
        self.directory = directory

    def __get__(self, instance, owner):
        slug = super(AutoTemplate, self).__get__(instance, owner)
        return u'advanced_reports/backoffice/{0}/{1}.html'.format(self.directory, slug)


class BackOfficeTab(object):
    """
    A tab that will be shown for a ``BackOfficeModel``. A ``BackOfficeModel`` can
    have some tabs associated with it. Each tab is represented by:

    *   ``slug``: The unique (inside a model) slug for a tab.
        Used for navigation in the URL.
    *   ``title``: A human title that will pe used as the caption for a tab.
    *   ``template``: The path to a template that must be rendered inside the tab.
        This template is given the ``instance`` context variable, which contains
        the current model instance.
    """

    slug = None
    title = None
    template = None
    permission = None

    def __init__(self, slug, title, template=None, permission=None, shadow=None, context=None):
        self.slug = slug
        self.title = title
        if template:
            self.template = template
        else:
            self.template = u'advanced_reports/backoffice/{0}/{1}.html'.format('tabs', slug)
        self.shadow = shadow
        self.permission = permission
        self.context = context or {}

    def get_serialized_meta(self):
        return {
            'slug': self.slug,
            'title': self.title,
            'shadow': self.shadow
        }

    def get_serialized(self, request, instance):
        context = {'instance': instance}
        context.update(self.context)
        template = self.template or 'advanced_reports/backoffice/default-tab.html'
        return {
            'slug': self.slug,
            'title': self.title,
            'template': render_to_string(template, context,
                                         context_instance=RequestContext(request)) + random_token(),
            'shadow': self.shadow
        }

    def __repr__(self):
        return 'BackOfficeTab(%r, %r, %r, permission=%r, shadow=%r, context=%r)' \
               % (self.slug, self.title, self.template, self.permission, self.shadow, self.context)


class BackOfficeModel(object):
    """
    The base class for a ``BackOfficeModel`` implementation. Implement this
    class and register your own Django models to a backoffice.
    """

    #: A unique slug to identify a model ('user', 'sim', ...)
    slug = AutoSlug(remove_suffix='Model')

    #: The actual Django model that is being used
    model = None

    #: (optional) The name of a ``ForeignKey`` that points to a parent object.
    #: If this parent is registered with the same backoffice, they are shown as
    #: a parent/child relation.
    parent_field = None

    #: (optional) The plural version of parent_field.
    parent_fields = None

    #: Verbose name for displaying purposes. Should be lowercase. E.g. 'user'
    verbose_name = None

    #: Plural verbose name for displaying purposes. Should be lowercase. E.g. 'users'
    verbose_name_plural = None

    #: A mapping of slugs to ``BackOfficeModel`` implementation instances
    #: to represent kinds of children. This property will be automatically filled
    #: when you register your parent and child models with a BackOfficeBase
    #: implementation instance.
    children = None
    child_to_accessor = None

    #: A mapping of slugs to ``BackOfficeModel`` implementation instances
    #: to represent kinds of parents. This property will be automatically filled
    #: when you register your parent and child models with a BackOfficeBase
    #: implementation instance.
    parents = None
    parent_to_accessor = None

    #: Include siblings of specified parent slug in serialization.
    siblings = None

    #: Define a priority that will be used for displaying purposes.
    #: Can also be used as a way to sort models by kind, if used uniquely.
    priority = 999

    #: For displaying purposes. Will probably change, so leave alone.
    has_header = True

    #: For displaying purposes. Will probably change, so leave alone.
    collapsed = True

    #: Whether we want to appear as a menu of children in the parent view
    show_in_parent = False

    #: An optional template path to render a header for a model instance
    #: above the tabs.
    header_template = None

    #: A tuple of ``BackOfficeTab`` instances. These tabs can contain templates
    #: which in turn can display extra information about this model.
    tabs = ()

    #: When you have overridden ``search_index`` and included child objects
    #: in your search index (e.g. comments on a user) you can define these
    #: child models to trigger a reindex when a child changes.
    #: Example: ``{Comment: (lambda c: c.content_object, User)}``.
    #: In the value of this dictionary you define a tuple with a function
    #: which knows how to find the parent object and a model type to make
    #: sure we are getting the right argument to our ``search_index``
    #: implementation.
    search_index_dependencies = None

    #: An optional string containing a permission which has to be checked
    #: against the request that queries this model. If the permission is
    #: not satisfied, the model will not show up in search results and
    #: will never display.
    permission = None

    def __init__(self):
        self.children = {}
        self.child_to_accessor = {}
        self.parents = {}
        self.parent_to_accessor = {}
        self.search_index_dependencies = {}

    def get_title(self, instance):
        """
        A textual representation of a model instance to be used as a title.
        """
        return unicode(instance)

    def serialize(self, instance):
        """
        Override this to add extra information to be exposed to the Angular
        side of your tab templates.
        """
        return {}

    def render_template(self, request, instance):
        if self.header_template:
            context = {
                'instance': instance
            }
            return render_to_string(self.header_template, context,
                                    context_instance=RequestContext(request))
        return u''

    def get_parent_model_slug_for_field(self, parent_field):
        return None

    def get_parent_accessor_to_myself(self, parent_field):
        parent_django_field = self.model._meta.get_field(parent_field)
        return parent_django_field.rel.related_name \
            or '%s_set' % self.model.__name__.lower()

    def get_parent_accessor_from_myself(self, parent_field):
        return parent_field

    def get_route(self, instance):
        return {'model': self.slug, 'id': instance.pk}

    def get_serialized(self, request, instance, children=False, parents=False, siblings=False, templates=False):
        serialized = {
            'id': instance.pk,
            'title': self.get_title(instance),
            'model': self.slug,
            'route': self.get_route(instance),
            'is_object': True,
            'meta': self.serialize_meta(request)
        }

        if templates:
            serialized['tabs'] = dict((t.slug, t.get_serialized(request, instance)) \
                                 for t in self.tabs \
                                 if check_permission(request, t.permission))
            serialized['header_template'] = self.render_template(request, instance) + random_token()

        if children:
            serialized['children'] = self.get_children(request, instance)

        if parents:
            serialized['parents'] = self.get_parents(request, instance)

        if self.siblings and siblings:
            parent_bo_model = self.parents[self.siblings]
            parent = self.get_parent_by_model(request, instance, parent_bo_model)
            serialized['siblings'] = parent_bo_model.get_serialized_children_by_model(request, parent, self, exclude_meta=True)

        serialized.update(self.serialize(instance))
        return serialized

    @property
    def parent_fields_list(self):
        return self.parent_fields or self.parent_field and [self.parent_field]

    def get_children_by_model(self, request, instance, bo_model):
        if not instance or not check_permission(request, bo_model.permission):
            return []
        return getattr(instance, self.child_to_accessor[bo_model.slug]).all()

    def get_serialized_children_by_model(self, request, instance, bo_model, exclude_meta=False):
        if not check_permission(request, bo_model.permission):
            return []
        children = self.get_children_by_model(request, instance, bo_model)
        serialized_children = [bo_model.get_serialized(request, c) for c in children]
        if bo_model.has_header and not exclude_meta:
            serialized = bo_model.serialize_meta(request)
            serialized['children'] = serialized_children
            return [serialized]
        else:
            return serialized_children

    def get_parent_by_model(self, request, instance, bo_model):
        if not check_permission(request, bo_model.permission):
            return None
        parent = getattr(instance, self.parent_to_accessor[bo_model.slug])
        if parent is None:
            return None
        if not isinstance(parent, bo_model.model):
            return None
        return parent

    def get_serialized_parent_by_model(self, request, instance, bo_model):
        if not check_permission(request, bo_model.permission):
            return None
        parent = self.get_parent_by_model(request, instance, bo_model)
        return bo_model.get_serialized(request, parent) if parent else None

    def serialize_meta(self, request):
        return {
            'slug': self.slug,
            'verbose_name': self.verbose_name or self.model._meta.verbose_name,
            'verbose_name_plural': self.verbose_name_plural or self.model._meta.verbose_name_plural,
            'has_header': self.has_header,
            'collapsed': self.collapsed,
            'show_in_parent': self.show_in_parent,
            'tabs': [t.get_serialized_meta() for t in self.tabs if check_permission(request, t.permission)],
            'is_meta': True
        }

    def get_children(self, request, instance):
        if not self.children:
            return ()
        child_models = sorted(self.children.values(), key=lambda m: m.priority)
        return sum((self.get_serialized_children_by_model(request, instance, m) for m in child_models), [])

    def get_parents(self, request, instance):
        if not self.parents:
            return ()
        parent_models = sorted(self.parents.values(), key=lambda m: m.priority)
        parents = (self.get_serialized_parent_by_model(request, instance, parent_model) for parent_model in parent_models)
        return [parent for parent in parents if parent]

    def reindex(self, instance, backoffice_instance):
        index_text = self.search_index(instance)
        index, created = SearchIndex.objects.get_or_create(backoffice_instance=backoffice_instance,
                                                           model_slug=self.slug,
                                                           model_id=instance.pk,
                                                           defaults={'to_index': index_text})
        if not created:
            index.to_index = index_text
            index.save()

    def delete_index(self, instance, backoffice_instance):
        try:
            index = SearchIndex.objects.get(backoffice_instance=backoffice_instance,
                                            model_slug=self.slug,
                                            model_id=instance.pk)
            index.delete()
        except ObjectDoesNotExist:
            pass

    def search_index(self, instance):
        return unicode(instance)


class ModelMixin(object):
    """
    This mixin implements support for BackOfficeModels.

    Requires that ``self.name`` is available.
    """

    #: A mapping between Django models and ``BackOfficeModel`` implementation
    #: instances.
    model_to_bo_model = None

    #: A mapping between the ``slug`` of a ``BackOfficeModel`` implementation
    #: and ``BackOfficeModel`` implementation instances.
    #: This is where slugs will be resolved to their actual ``BackOfficeModel``
    #: instance.
    slug_to_bo_model = None

    #: A mapping between Django models and lists of ``BackOfficeModel``
    #: instances that depend on the Django models for their ``search_index``.
    search_index_dependency_to_dependants = defaultdict(lambda: [])

    def _reindex_model_signal_handler(self, sender, **kwargs):
        """
        Signal handler for re-indexing models. If the sender is a dependency
        for another model, the other model will also be indexed.
        """
        bo_model = self.get_model(model=sender)
        instance = kwargs.get('instance')

        # Just do a simple reindex
        if bo_model:
            bo_model.reindex(instance, self.name)

        # Reindex search_index dependants if there are any.
        if sender in self.search_index_dependency_to_dependants:
            dependants = self.search_index_dependency_to_dependants[sender]
            for dependant in dependants:
                get_dependant_instance, expected_dependant_model = dependant.search_index_dependencies[sender]
                dependant_instance = get_dependant_instance(instance)
                if isinstance(dependant_instance, expected_dependant_model):
                    dependant.reindex(dependant_instance, self.name)

    def _delete_index_signal_handler(self, sender, **kwargs):
        """
        Signal handler for deleting search indexes for models that will
        be deleted..
        """
        bo_model = self.get_model(model=sender)
        instance = kwargs.get('instance')

        if bo_model:
            bo_model.delete_index(instance, self.name)

    def register_model(self, bo_model):
        """
        Register a ``BackOfficeModel`` implementation class with this
        backoffice.

        :param bo_model: a ``BackOfficeModel`` implementation class
        """
        if not self.model_to_bo_model:
            self.model_to_bo_model = {}
        if not self.slug_to_bo_model:
            self.slug_to_bo_model = {}

        if not bo_model.slug in self.slug_to_bo_model:
            bo_model_instance = bo_model()
            self.slug_to_bo_model[bo_model.slug] = bo_model_instance
            self.model_to_bo_model[bo_model.model] = bo_model_instance

            # Connect signals to listen for changes to this model and their
            # ``search_index_dependencies``.
            post_save.connect(self._reindex_model_signal_handler, sender=bo_model.model)
            post_delete.connect(self._delete_index_signal_handler, sender=bo_model.model)

            for dependency in bo_model_instance.search_index_dependencies:
                self.search_index_dependency_to_dependants[dependency].append(bo_model_instance)
                post_save.connect(self._reindex_model_signal_handler, sender=dependency)
                post_delete.connect(self._reindex_model_signal_handler, sender=dependency)

        for bo_model_instance in self.model_to_bo_model.values():
            self.link_relationship(bo_model_instance)

    def get_model(self, slug=None, model=None):
        """
        Gets a registered ``BackOfficeModel`` implementation instance either
        by slug or Django model.
        """
        if slug:
            return self.slug_to_bo_model.get(slug, None)
        else:
            return self.model_to_bo_model.get(model, None)

    def serialize_model_instance(self, request, instance, include_children=False):
        """
        Serializes an Django model instance that is registered with this
        backoffice using their ``BackOfficeModel`` implementation.

        :param instance: a Django model instance
        :param include_children: whether to include children in the
        serialization
        :return: a simple Python object that is JSON serializable
        """
        bo_model = self.get_model(model=type(instance))
        if not check_permission(request, bo_model.permission):
            return None
        if include_children:
            return bo_model.get_serialized(request, instance, children=True)
        else:
            return bo_model.get_serialized(request, instance)

    def serialize_model_instances(self, request, instances, include_children=False):
        """
        Serializes multiple Django model instances using
        ``serialize_model_instance``.

        :param instances: an iterable containing registered Django model
        instances registered with this backoffice.
        :param include_children: whether to include children in the
        serialization
        :return: a list of simple Python objects that are JSON serializable.
        """
        unfiltered = (self.serialize_model_instance(request, i, include_children=include_children) \
                      for i in instances)
        return [s for s in unfiltered if s]

    def link_relationship(self, bo_model):
        """
        Links parent ``BackOfficeModel`` implementations with their parents
        and children. The parent will get a ``children`` attribute containing
        a mapping between the children slugs and child ``BackOfficeModel``
        implementations.

        :param bo_model: a child ``BackOfficeModel`` implementation instance
        that wants to link with its parent.
        """
        parent_fields = bo_model.parent_fields_list

        if parent_fields:
            for parent_field in parent_fields:
                parent_model_slug = bo_model.get_parent_model_slug_for_field(parent_field)
                if parent_model_slug:
                    parent_bo_model = self.get_model(slug=parent_model_slug)
                else:
                    parent_model = bo_model.model._meta.get_field(parent_field).rel.to
                    parent_bo_model = self.get_model(model=parent_model)
                if parent_bo_model:
                    if not bo_model.slug in parent_bo_model.children:
                        parent_bo_model.children[bo_model.slug] = bo_model
                        parent_bo_model.child_to_accessor[bo_model.slug] = bo_model.get_parent_accessor_to_myself(parent_field)
                    if not parent_bo_model.slug in bo_model.parents:
                        bo_model.parents[parent_bo_model.slug] = parent_bo_model
                        bo_model.parent_to_accessor[parent_bo_model.slug] = bo_model.get_parent_accessor_from_myself(parent_field)

    def api_get_model(self, request):
        """
        API call to retrieve a serialized Django model instance using his
        ``BackOfficeModel`` implementation.

        :param request: ``request.view_params with 'model_slug' and 'pk'``
        :return: a serialized Django model instance using his
        ``BackOfficeModel`` implementation.
        """
        model_slug = request.view_params.get('model_slug')
        pk = request.view_params.get('pk')
        bo_model = self.get_model(slug=model_slug)
        if not check_permission(request, bo_model.permission):
            return HttpResponse(u'You are not allowed to view this page.', status=403)
        obj = bo_model.model.objects.get(pk=pk)
        serialized = bo_model.get_serialized(request, obj, children=True, parents=True, siblings=True, templates=True)
        return serialized

    def api_post_model_action(self, request):
        """
        Performs an action to a given view.

        :param request: ``request.view_params with 'method', 'params' and 'view_params'``
        :return: the results of the called view method.
        """
        method = request.view_params.get('method')
        action_params = request.view_params.get('params')
        model_slug = request.view_params.get('model')

        # Attach the actual view_params and action_params to the request.
        request.action_params = action_params

        bo_model = self.get_model(slug=model_slug)

        # Prevent posting to this view if not allowed to access this view
        if not check_permission(request, bo_model.permission):
            return HttpResponse(u'You are not allowed to post data to the model "%s".' % bo_model.slug, status=403)

        fn = getattr(bo_model, method, None)
        if not fn:
            raise Http404(u'Cannot find method %s on model %s' % (method, bo_model.slug))
        return fn(request)


class SearchMixin(object):
    """
    This mixin implements support for searching through BackOfficeModels.
    """

    def serialize_search_result(self, request, index):
        """
        Transforms a ``SearchIndex`` instance to a serialized ``BackOfficeModel``
        including metadata.
        """
        bo_model = self.get_model(slug=index.model_slug)
        if bo_model is None or not check_permission(request, bo_model.permission):
            return None
        try:
            instance = bo_model.model.objects.get(pk=index.model_id)
        except ObjectDoesNotExist:
            return None
        serialized = bo_model.get_serialized(request, instance)
        return serialized

    def count_by_model(self, request, indices):
        """
        Given an iterable of ``SearchIndex`` instances, return a report of
        counts by ``model_slug``, and their serialized ``BackOfficeModel``
        metadata.

        :param indices: an iterable of ``SearchIndex`` instances
        :returns: ``[{'meta': {'slug': 'user', ...}, 'count': 15}, ...]``
        """
        model_counts = defaultdict(lambda: 0)
        for index in indices:
            model_counts[index.model_slug] += 1
        model_counts = [[self.get_model(mc[0]), mc[1]] for mc in model_counts.items()]
        model_counts = [mc for mc in model_counts if mc[0] is not None]
        model_counts.sort(key=lambda x: x[0].priority)
        model_counts = [{'meta': mc[0].serialize_meta(request),
                         'count': mc[1]
                        } for mc in model_counts if check_permission(request, mc[0].permission)]
        return model_counts

    def serialize_search_results(self, request, indices):
        """
        Transforms ``SearchIndex`` instances to a list of serialized
        ``BackOfficeModel`` including metadata.
        """
        serialized = (self.serialize_search_result(request, i) \
                      for i in indices \
                      if i.model_slug in self.slug_to_bo_model)
        return [s for s in serialized if s]

    def search(self, request, query, filter_on_model_slug=None, page=1, page_size=20, include_counts=True):
        """
        Performs a search on Django models registered with this backoffice.

        :param query: The search query.
        :param filter_on_model_slug: (optional) a ``BackOfficeModel`` slug
        to filter on
        :param page: (optional) the page of the search results to show. By
        default 1.
        :param page_size: (optional) the page size. By default 20.
        :param include_counts: (optional) whether to include the counts.
        :return: ``{'results': [self.serialize_search_results()], 'model_counts': [self.count_by_model()]}``
        """
        if query == u'':
            return ()

        if not DB_IS_POSTGRES:
            all_indices = SearchIndex.objects.filter(to_index__icontains=query, backoffice_instance=self.name)
        else:
            ts_query = convert_to_raw_tsquery(query)
            all_indices = SearchIndex.objects.search(ts_query, raw=True).filter(backoffice_instance=self.name)

        model_counts = self.count_by_model(request, all_indices) if include_counts else []

        if filter_on_model_slug:
            all_indices = all_indices.filter(model_slug=filter_on_model_slug)

        indices = all_indices[(page-1)*page_size:page*page_size]

        return {
            'results': self.serialize_search_results(request, indices),
            'model_counts': model_counts,
            'query': query
        }

    def reindex_all_models(self):
        SearchIndex.objects.all().delete()
        for bo_model in self.slug_to_bo_model.values():
            for obj in bo_model.model.objects.all():
                bo_model.reindex(obj, self.name)

    def api_get_search(self, request):
        """
        API call implementation for ``self.search()``. It passes on ``q``,
        ``page`` and ``filter_model`` from the ``request.view_params``.

        :param request: a HttpRequest containing ``view_params``.
        :return: the results of ``self.search()``.
        """
        page = int(request.view_params.get('page', '1'))
        filter_model = request.view_params.get('filter_model')
        q = request.view_params.get('q')
        return self.search(request, q, page=page, filter_on_model_slug=filter_model)

    def api_get_search_preview(self, request):
        """
        API call to perform a search preview. It passes on ``q``, ``page``
        and ``filter_model``. It uses a ``page_size`` of 5.

        :param request: a HttpRequest containing ``view_params``.
        :return: the results of ``self.search(page_size=5)``.
        """
        filter_model = request.view_params.get('filter_model')
        q = request.view_params.get('q')
        return self.search(request, q, page_size=10, filter_on_model_slug=filter_model,
                           include_counts=False)

    def api_get_reindex(self, request):
        if request.user.is_authenticated():
            if request.user.is_superuser:
                self.reindex_all_models()
                return JSONResponse('OK')
        return HttpResponseForbidden('Forbidden')


class BackOfficeView(object):
    """
    A ``BackOfficeView`` is a combination of Python/Django code with the ``view`` AngularJS Directive.
    When registered to a ``BackOfficeBase`` subclass instance, this will be shown by the ``view`` directive
    when called with the correct ``slug``.
    """

    #: The slug of the view. If not explicitly filled in, the "View" suffix will be stripped off the
    #: subclass name and the remaining prefix will be converted to lowercase. For example, a
    #: ``UserView`` subclass will get automatically the slug ``'user'``.
    slug = AutoSlug(remove_suffix='View')

    #: The template for the view. If not explicitly filled in, it will default to, for example when the
    #: subclass has the name ``UserView``, ``'advanced_reports/backoffice/views/user.html'``.
    template = AutoTemplate('views', remove_suffix='View')

    #: When filled in, this permission will be checked against the currently logged in user. If
    #: the user does not have this permission, nothing of this view will be exposed to this user.
    permission = None

    def serialize(self, content, extra_context=None):
        """
        Serializes the given content to a format that is easy to encode to JSON.
        """
        if isinstance(content, str):
            content = content.decode('utf-8')

        context = {
            'slug': self.slug,
            'content': content + random_token(),
        }
        if extra_context:
            context.update(extra_context)
        return context

    def serialize_view_result(self, request, result):
        content, extra_context = result, {}
        if isinstance(result, tuple):
            content, extra_context = result
        if isinstance(content, HttpResponse):
            if hasattr(content, 'render'):
                content.render()
            content = content.content
        return self.serialize(content, extra_context)

    def get_serialized(self, request):
        return self.serialize_view_result(request, self.get(request))

    def get_serialized_post(self, request):
        return self.serialize_view_result(request, self.post(request))

    def get_context(self, request):
        context = {}
        context.update(self.get_extra_context(request))
        context.update({'view_slug': self.slug})
        return context

    def get_extra_context(self, request):
        return {}

    def get(self, request):
        return render(request, self.template, self.get_context(request))

    def post(self, request):
        raise NotImplementedError

    def FOO(self, request):
        raise NotImplementedError


class ViewMixin(object):
    """
    This mixin implements support for BackOfficeViews.
    """

    #: The internal view registration dict
    slug_to_bo_view = None

    def register_view(self, bo_view):
        """
        Registers a ``BackOfficeView`` implementation class with this
        backoffice.

        :param bo_view: a ``BackOfficeView`` implementation class
        """
        if not self.slug_to_bo_view:
            self.slug_to_bo_view = {}

        if not bo_view.slug in self.slug_to_bo_view:
            bo_view_instance = bo_view()
            self.slug_to_bo_view[bo_view.slug] = bo_view_instance

    def get_view(self, slug):
        """
        Gets a registered ``BackOfficeView`` implementation instance by slug.
        """
        return self.slug_to_bo_view and self.slug_to_bo_view.get(slug, None)

    def api_get_view(self, request):
        """
        Retrieves a ``BackOfficeView`` implementation based on the given slug
        and calls the ``get_serialized()`` method on it, whose contents are
        returned.

        :param request: ``request.view_params with 'view_slug'``
        :return: a serialized view content
        """
        slug = request.view_params.get('view_slug')
        bo_view = self.get_view(slug)

        if not bo_view:
            return {'content': u'<span class="text-danger">View with slug "%s" does not exist. Are you sure you have registered it?</span>' % slug}

        # Just hide the view if the viewing of it is not permitted
        if not check_permission(request, bo_view.permission):
            return {'content': u''}

        return bo_view.get_serialized(request)

    def api_post_view(self, request):
        """
        Retrieves a ``BackOfficeView`` implementation based on the given slug
        and calls the ``get_serialized_post()`` method on it, whose contents are
        returned.

        :param request: ``request.view_params with 'view_slug'``
        :return: a serialized view content
        """
        bo_view = self.get_view(request.view_params.get('view_slug'))

        # Prevent posting to this view if not allowed to access this view
        if not check_permission(request, bo_view.permission):
            return HttpResponse(u'You are not allowed to post data to the view "%s".' % bo_view.slug, status=403)

        return bo_view.get_serialized_post(request)

    def api_post_view_action(self, request):
        """
        Performs an action to a given view.

        :param request: ``request.view_params with 'method', 'params' and 'view_params'``
        :return: the results of the called view method.
        """
        method = request.view_params.get('method')
        action_params = request.view_params.get('params')
        view_params = request.view_params.get('view_params')

        # Attach the actual view_params and action_params to the request.
        request.action_params = action_params
        request.view_params = view_params

        bo_view = self.get_view(view_params.get('view_slug'))

        # Prevent posting to this view if not allowed to access this view
        if not check_permission(request, bo_view.permission):
            return HttpResponse(u'You are not allowed to post data to the view "%s".' % bo_view.slug, status=403)

        fn = getattr(bo_view, method, None)
        if not fn:
            raise Http404(u'Cannot find method %s on view %s' % (method, bo_view.slug))
        return fn(request)

    def api_get_view_view(self, request):
        """
        Calls a regular Django view on the given BackOffice view.

        :param request: ``request.view_params with 'method' and 'view_slug'``
        :return: the HttpResponse from the BackOffice method with the name inside 'method'
        """
        method = request.view_params.get('method')
        view_slug = request.view_params.get('view_slug')

        bo_view = self.get_view(view_slug)

        # Prevent accessing this view if not allowed
        if not check_permission(request, bo_view.permission):
            return HttpResponse(u'You are not allowed to view data from the view "%s".' % bo_view.slug, status=403)

        fn = getattr(bo_view, method, None)
        if not fn:
            raise Http404(u'Cannot find method %s on view %s' % (method, bo_view.slug))
        return fn(request)

    def api_post_view_view(self, request):
        """
        Alias to ``api_get_view_view`` to support POST requests as well.
        """
        return self.api_get_view_view(request)


class BackOfficeBase(ViewMixin, SearchMixin, ModelMixin):
    """
    The base class of a Back Office application. Inherit from this class and
    define your own custom backoffice!
    """

    #: The title of the BackOffice application. This is used on pages that
    #: inherit from the ``model_template``.
    title = 'Untitled Backoffice'

    #: The template file that will be used to show models. It has to define
    #: an ng-view inside of it. Please extend the default template if you
    #: want to specify your own navigation (which you will probably want
    #: to do) and put your own template here.
    model_template = 'advanced_reports/backoffice/model-base.html'

    #: The template file that will be used to show the login form.
    login_template = 'advanced_reports/backoffice/login.html'

    #: The template file that will be used as the logout confirmation page.
    logout_template = 'advanced_reports/backoffice/logout.html'

    def __init__(self, name='backoffice', app_name='backoffice'):
        """
        Constructor for a BackOfficeBase implementation.

        :param name: the instance name for this implementation used for
        Django url namespacing.
        :param app_name: the app name for this implementation. In most cases
        can be just left alone.
        :return: a BackOfficeBase implementation instance.
        """
        self.name = name
        self.app_name = app_name

    def define_urls(self):
        """
        Implement this to add some custom url patterns to a backoffice.
        They will automatically get the proper namespacing. For example,
        if your backoffice is called "helpdesk" and your url is called
        "stats", you must use ``{% url 'helpdesk:stats' %}`` in your
        templates.

        :return: a tuple/list of url patterns (not using ``patterns()``!)
        """
        return ()

    @property
    def urls(self):
        """
        The actual url patterns for this backoffice. You can include
        these in your main urlconf.

        :return: url patterns
        """
        return patterns('',
                        url(r'^$', self.decorate(self.home), name='home'),
                        url(r'^logout/$', self.logout, name='logout'),
                        url(r'^api/(?P<method>[^/]+)/$', self.decorate(self.api), name='api'),
                        url(r'^api/$', self.decorate(self.api), name='api_home'),
                        url(r'^login/as/(?P<user_id>\d+)/$', self.decorate(self.login_as), name='login_as'),
                        url(r'^end/login/as/$', self.end_login_as, name='end_login_as'),
                        url(r'^handle/$', self.handle, name='handle'),
                        *self.define_urls()
        ), self.app_name, self.name

    def decorate(self, view):
        """
        Implement this to add some custom decoration for the internal
        Django views for this backoffice.

        :param view: the view to decorate
        :return: the decorated view
        """
        # TODO: add permissions
        return staff_member_required(self)(view)

    def default_context(self):
        """
        Default template context values for internal backoffice page.

        :return: context dict
        """
        return {'backoffice': self,
                'api_url': reverse(self.name + ':api_home', current_app=self.app_name),
                'root_url': reverse(self.name + ':home', current_app=self.app_name)}

    @never_cache
    def logout(self, request, *args, **kwargs):
        """
        Calls ``django.contrib.auth.views.logout`` with a custom template and
        extra context.
        """
        from django.contrib.auth.views import logout
        kwargs['template_name'] = self.logout_template
        kwargs['extra_context'] = self.default_context()
        return logout(request, *args, **kwargs)

    def home(self, request):
        """
        The main Django view for this backoffice.
        """
        return render_to_response(self.model_template,
                                  self.default_context(),
                                  context_instance=RequestContext(request))

    # Impersonation
    def login_as(self, request, user_id):
        me_id = request.session.get('impersonation_original_user', None)
        me = User.objects.get(pk=me_id) if me_id else request.user
        user = get_object_or_404(User, pk=user_id)
        user.backend = 'django.contrib.auth.backends.ModelBackend'
        login(request, user)
        request.session['impersonation_original_user'] = me.pk
        request.session['impersonating'] = True
        return redirect(settings.LOGIN_REDIRECT_URL)

    def end_login_as(self, request):
        if 'impersonation_original_user' in request.session:
            me_id = request.session['impersonation_original_user']
            me = get_object_or_404(User, pk=me_id)
            me.backend = 'django.contrib.auth.backends.ModelBackend'
            del request.session['impersonation_original_user']
            del request.session['impersonating']
            login(request, me)
            return redirect(reverse('backoffice:home', current_app=self.name))
        return redirect(settings.LOGIN_REDIRECT_URL)

    def api(self, request, method=None):
        """
        A very simple REST-like API implementation which just passes requests
        on to the right functions.

        An instance of ``advanced_reports.backoffice.api_utils.ViewRequestParameters``
        will be assigned to the ``request`` as ``view_params``.

        :param request: A HTTP request where ``view_params`` will be attached to.
        :param method: The actual instance method of this backoffice class that will
        be called.
        :return: a JSONResponse
        """
        if not method:
            return JSONResponse(None)

        request.view_params = ViewRequestParameters(request)

        fn = getattr(self, 'api_%s_%s' % (request.method.lower(), method), None)
        if fn is None:
            raise Http404

        response = fn(request)

        if isinstance(response, HttpResponse):
            return response

        msgs = [m.__dict__ for m in messages.get_messages(request)]
        return JSONResponse({'messages': msgs, 'response_data': response})

    def handle(self, request):
        raise Http404



from django.conf.urls import url
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse

from advanced_reports.backoffice.base import BackOfficeBase, BackOfficeModel, BackOfficeTab, BackOfficeView
from advanced_reports.backoffice.contrib.views import AdvancedReportView
import advanced_reports


from advanced_reports.backoffice.examples.reports import UserReport, UserForm
from advanced_reports.backoffice.examples.views import SimpleView


User = get_user_model()


class TestBackOffice(BackOfficeBase):
    title = 'Test Backoffice'

    def __init__(self):
        super(TestBackOffice, self).__init__(name='test_backoffice')

    def define_urls(self):
        return (url(r'^page/$', self.page, name='page'),)

    def page(self, request):
        return TemplateResponse(request, 'advanced_reports/backoffice/tests/page.html', {'backoffice': self})


class UserModel(BackOfficeModel):
    model = User
    verbose_name = u'user'
    verbose_name_plural = u'users'

    tabs = (BackOfficeTab('details', u'Details', 'advanced_reports/backoffice/tests/user-details.html'),)

    def get_title(self, instance):
        return u'%s (%s)' % (instance.username, instance.get_full_name())

    def search_index(self, instance):
        return u' '.join((instance.username, instance.get_full_name(), instance.email))


class UserView(BackOfficeView):
    def render(self, request, user, form):
        return TemplateResponse(request, 'advanced_reports/backoffice/tests/user-form.html', {'user': user, 'form': form})

    def get(self, request):
        user = get_object_or_404(User, pk=request.view_params.get('pk'))
        form = UserForm(instance=user)
        return self.render(request, user, form)

    def post(self, request):
        user = get_object_or_404(User, pk=request.view_params.get('pk'))
        form = UserForm(request.POST, instance=user)
        if form.is_valid():
            form.save()
            messages.success(request, u'Successfully saved user details! Try a search now...')
        return self.render(request, user, form)

    def update(self, request):
        data = request.view_params.get('data')
        user = get_object_or_404(User, pk=request.view_params.get('pk'))
        user.first_name = data['first_name']
        user.last_name = data['last_name']
        user.save()
        return data


test_backoffice = TestBackOffice()
test_backoffice.register_model(UserModel)
test_backoffice.register_view(AdvancedReportView)
test_backoffice.register_view(UserView)
test_backoffice.register_view(SimpleView)

advanced_reports.register(UserReport)



import importlib

from django.core.management.base import BaseCommand
from advanced_reports.backoffice.models import SearchIndex


class Command(BaseCommand):
    help = 'Cleans the search index. Argument: Python path to backoffice instance.'

    option_list = BaseCommand.option_list

    def handle(self, *args, **options):
        backoffice_path = args[0]
        module, obj = backoffice_path.rsplit('.', 1)
        backoffice_module = importlib.import_module(module)
        backoffice = getattr(backoffice_module, obj)

        count = 0
        models_not_found = {}

        for si in SearchIndex.objects.filter(backoffice_instance=backoffice.name):
            bo_model = backoffice.get_model(slug=si.model_slug)
            if not bo_model:
                models_not_found[si.model_slug] = True
            else:
                if not bo_model.model.objects.filter(pk=si.model_id).exists():
                    si.delete()
                    count += 1
        if models_not_found:
            self.stdout.write(u'Following models were not found: %s\n' % ', '.join(models_not_found))
        self.stdout.write(u'Done deleting %d unnecessary search indexes!\n' % count)



#!/Users/jef/env/mv-poland/bin/python
"""PILdriver, an image-processing calculator using PIL.

An instance of class PILDriver is essentially a software stack machine
(Polish-notation interpreter) for sequencing PIL image
transformations.  The state of the instance is the interpreter stack.

The only method one will normally invoke after initialization is the
`execute' method.  This takes an argument list of tokens, pushes them
onto the instance's stack, and then tries to clear the stack by
successive evaluation of PILdriver operators.  Any part of the stack
not cleaned off persists and is part of the evaluation context for
the next call of the execute method.

PILDriver doesn't catch any exceptions, on the theory that these
are actually diagnostic information that should be interpreted by
the calling code.

When called as a script, the command-line arguments are passed to
a PILDriver instance.  If there are no command-line arguments, the
module runs an interactive interpreter, each line of which is split into
space-separated tokens and passed to the execute method.

In the method descriptions below, a first line beginning with the string
`usage:' means this method can be invoked with the token that follows
it.  Following <>-enclosed arguments describe how the method interprets
the entries on the stack.  Each argument specification begins with a
type specification: either `int', `float', `string', or `image'.

All operations consume their arguments off the stack (use `dup' to
keep copies around).  Use `verbose 1' to see the stack state displayed
before each operation.

Usage examples:

    `show crop 0 0 200 300 open test.png' loads test.png, crops out a portion
of its upper-left-hand corner and displays the cropped portion.

    `save rotated.png rotate 30 open test.tiff' loads test.tiff, rotates it
30 degrees, and saves the result as rotated.png (in PNG format).
"""
# by Eric S. Raymond <esr@thyrsus.com>
# $Id$

# TO DO:
# 1. Add PILFont capabilities, once that's documented.
# 2. Add PILDraw operations.
# 3. Add support for composing and decomposing multiple-image files.
#

from PIL import Image
import string

class PILDriver:

    verbose = 0

    def do_verbose(self):
        """usage: verbose <int:num>

        Set verbosity flag from top of stack.
        """
        self.verbose = self.do_pop()

    # The evaluation stack (internal only)

    stack = []          # Stack of pending operations

    def push(self, item):
        "Push an argument onto the evaluation stack."
        self.stack = [item] + self.stack

    def top(self):
        "Return the top-of-stack element."
        return self.stack[0]

    # Stack manipulation (callable)

    def do_clear(self):
        """usage: clear

        Clear the stack.
        """
        self.stack = []

    def do_pop(self):
        """usage: pop

        Discard the top element on the stack.
        """
        top = self.stack[0]
        self.stack = self.stack[1:]
        return top

    def do_dup(self):
        """usage: dup

        Duplicate the top-of-stack item.
        """
        if hasattr(self, 'format'):     # If it's an image, do a real copy
            dup = self.stack[0].copy()
        else:
            dup = self.stack[0]
        self.stack = [dup] + self.stack

    def do_swap(self):
        """usage: swap

        Swap the top-of-stack item with the next one down.
        """
        self.stack = [self.stack[1], self.stack[0]] + self.stack[2:]

    # Image module functions (callable)

    def do_new(self):
        """usage: new <int:xsize> <int:ysize> <int:color>:

        Create and push a greyscale image of given size and color.
        """
        xsize = int(self.do_pop())
        ysize = int(self.do_pop())
        color = int(self.do_pop())
        self.push(Image.new("L", (xsize, ysize), color))

    def do_open(self):
        """usage: open <string:filename>

        Open the indicated image, read it, push the image on the stack.
        """
        self.push(Image.open(self.do_pop()))

    def do_blend(self):
        """usage: blend <image:pic1> <image:pic2> <float:alpha>

        Replace two images and an alpha with the blended image.
        """
        image1 = self.do_pop()
        image2 = self.do_pop()
        alpha = float(self.do_pop())
        self.push(Image.blend(image1, image2, alpha))

    def do_composite(self):
        """usage: composite <image:pic1> <image:pic2> <image:mask>

        Replace two images and a mask with their composite.
        """
        image1 = self.do_pop()
        image2 = self.do_pop()
        mask = self.do_pop()
        self.push(Image.composite(image1, image2, mask))

    def do_merge(self):
        """usage: merge <string:mode> <image:pic1> [<image:pic2> [<image:pic3> [<image:pic4>]]]

        Merge top-of stack images in a way described by the mode.
        """
        mode = self.do_pop()
        bandlist = []
        for band in mode:
            bandlist.append(self.do_pop())
        self.push(Image.merge(mode, bandlist))

    # Image class methods

    def do_convert(self):
        """usage: convert <string:mode> <image:pic1>

        Convert the top image to the given mode.
        """
        mode = self.do_pop()
        image = self.do_pop()
        self.push(image.convert(mode))

    def do_copy(self):
        """usage: copy <image:pic1>

        Make and push a true copy of the top image.
        """
        self.dup()

    def do_crop(self):
        """usage: crop <int:left> <int:upper> <int:right> <int:lower> <image:pic1>

        Crop and push a rectangular region from the current image.
        """
        left = int(self.do_pop())
        upper = int(self.do_pop())
        right = int(self.do_pop())
        lower = int(self.do_pop())
        image = self.do_pop()
        self.push(image.crop((left, upper, right, lower)))

    def do_draft(self):
        """usage: draft <string:mode> <int:xsize> <int:ysize>

        Configure the loader for a given mode and size.
        """
        mode = self.do_pop()
        xsize = int(self.do_pop())
        ysize = int(self.do_pop())
        self.push(self.draft(mode, (xsize, ysize)))

    def do_filter(self):
        """usage: filter <string:filtername> <image:pic1>

        Process the top image with the given filter.
        """
        import ImageFilter
        filter = eval("ImageFilter." + string.upper(self.do_pop()))
        image = self.do_pop()
        self.push(image.filter(filter))

    def do_getbbox(self):
        """usage: getbbox

        Push left, upper, right, and lower pixel coordinates of the top image.
        """
        bounding_box = self.do_pop().getbbox()
        self.push(bounding_box[3])
        self.push(bounding_box[2])
        self.push(bounding_box[1])
        self.push(bounding_box[0])

    def do_getextrema(self):
        """usage: extrema

        Push minimum and maximum pixel values of the top image.
        """
        extrema = self.do_pop().extrema()
        self.push(extrema[1])
        self.push(extrema[0])

    def do_offset(self):
        """usage: offset <int:xoffset> <int:yoffset> <image:pic1>

        Offset the pixels in the top image.
        """
        xoff = int(self.do_pop())
        yoff = int(self.do_pop())
        image = self.do_pop()
        self.push(image.offset(xoff, yoff))

    def do_paste(self):
        """usage: paste <image:figure> <int:xoffset> <int:yoffset> <image:ground>

        Paste figure image into ground with upper left at given offsets.
        """
        figure = self.do_pop()
        xoff = int(self.do_pop())
        yoff = int(self.do_pop())
        ground = self.do_pop()
        if figure.mode == "RGBA":
            ground.paste(figure, (xoff, yoff), figure)
        else:
            ground.paste(figure, (xoff, yoff))
        self.push(ground)

    def do_resize(self):
        """usage: resize <int:xsize> <int:ysize> <image:pic1>

        Resize the top image.
        """
        ysize = int(self.do_pop())
        xsize = int(self.do_pop())
        image = self.do_pop()
        self.push(image.resize((xsize, ysize)))

    def do_rotate(self):
        """usage: rotate <int:angle> <image:pic1>

        Rotate image through a given angle
        """
        angle = int(self.do_pop())
        image = self.do_pop()
        self.push(image.rotate(angle))

    def do_save(self):
        """usage: save <string:filename> <image:pic1>

        Save image with default options.
        """
        filename = self.do_pop()
        image = self.do_pop()
        image.save(filename)

    def do_save2(self):
        """usage: save2 <string:filename> <string:options> <image:pic1>

        Save image with specified options.
        """
        filename = self.do_pop()
        options = self.do_pop()
        image = self.do_pop()
        image.save(filename, None, options)

    def do_show(self):
        """usage: show <image:pic1>

        Display and pop the top image.
        """
        self.do_pop().show()

    def do_thumbnail(self):
        """usage: thumbnail <int:xsize> <int:ysize> <image:pic1>

        Modify the top image in the stack to contain a thumbnail of itself.
        """
        ysize = int(self.do_pop())
        xsize = int(self.do_pop())
        self.top().thumbnail((xsize, ysize))

    def do_transpose(self):
        """usage: transpose <string:operator> <image:pic1>

        Transpose the top image.
        """
        transpose = string.upper(self.do_pop())
        image = self.do_pop()
        self.push(image.transpose(transpose))

    # Image attributes

    def do_format(self):
        """usage: format <image:pic1>

        Push the format of the top image onto the stack.
        """
        self.push(self.pop().format)

    def do_mode(self):
        """usage: mode <image:pic1>

        Push the mode of the top image onto the stack.
        """
        self.push(self.pop().mode)

    def do_size(self):
        """usage: size <image:pic1>

        Push the image size on the stack as (y, x).
        """
        size = self.pop().size
        self.push(size[0])
        self.push(size[1])

    # ImageChops operations

    def do_invert(self):
        """usage: invert <image:pic1>

        Invert the top image.
        """
        import ImageChops
        self.push(ImageChops.invert(self.do_pop()))

    def do_lighter(self):
        """usage: lighter <image:pic1> <image:pic2>

        Pop the two top images, push an image of the lighter pixels of both.
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        self.push(ImageChops.lighter(image1, image2))

    def do_darker(self):
        """usage: darker <image:pic1> <image:pic2>

        Pop the two top images, push an image of the darker pixels of both.
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        self.push(ImageChops.darker(image1, image2))

    def do_difference(self):
        """usage: difference <image:pic1> <image:pic2>

        Pop the two top images, push the difference image
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        self.push(ImageChops.difference(image1, image2))

    def do_multiply(self):
        """usage: multiply <image:pic1> <image:pic2>

        Pop the two top images, push the multiplication image.
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        self.push(ImageChops.multiply(image1, image2))

    def do_screen(self):
        """usage: screen <image:pic1> <image:pic2>

        Pop the two top images, superimpose their inverted versions.
        """
        import ImageChops
        image2 = self.do_pop()
        image1 = self.do_pop()
        self.push(ImageChops.screen(image1, image2))

    def do_add(self):
        """usage: add <image:pic1> <image:pic2> <int:offset> <float:scale>

        Pop the two top images, produce the scaled sum with offset.
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        scale = float(self.do_pop())
        offset = int(self.do_pop())
        self.push(ImageChops.add(image1, image2, scale, offset))

    def do_subtract(self):
        """usage: subtract <image:pic1> <image:pic2> <int:offset> <float:scale>

        Pop the two top images, produce the scaled difference with offset.
        """
        import ImageChops
        image1 = self.do_pop()
        image2 = self.do_pop()
        scale = float(self.do_pop())
        offset = int(self.do_pop())
        self.push(ImageChops.subtract(image1, image2, scale, offset))

    # ImageEnhance classes

    def do_color(self):
        """usage: color <image:pic1>

        Enhance color in the top image.
        """
        import ImageEnhance
        factor = float(self.do_pop())
        image = self.do_pop()
        enhancer = ImageEnhance.Color(image)
        self.push(enhancer.enhance(factor))

    def do_contrast(self):
        """usage: contrast <image:pic1>

        Enhance contrast in the top image.
        """
        import ImageEnhance
        factor = float(self.do_pop())
        image = self.do_pop()
        enhancer = ImageEnhance.Color(image)
        self.push(enhancer.enhance(factor))

    def do_brightness(self):
        """usage: brightness <image:pic1>

        Enhance brightness in the top image.
        """
        import ImageEnhance
        factor = float(self.do_pop())
        image = self.do_pop()
        enhancer = ImageEnhance.Color(image)
        self.push(enhancer.enhance(factor))

    def do_sharpness(self):
        """usage: sharpness <image:pic1>

        Enhance sharpness in the top image.
        """
        import ImageEnhance
        factor = float(self.do_pop())
        image = self.do_pop()
        enhancer = ImageEnhance.Color(image)
        self.push(enhancer.enhance(factor))

    # The interpreter loop

    def execute(self, list):
        "Interpret a list of PILDriver commands."
        list.reverse()
        while len(list) > 0:
            self.push(list[0])
            list = list[1:]
            if self.verbose:
                print "Stack: " + `self.stack`
            top = self.top()
            if type(top) != type(""):
                continue;
            funcname = "do_" + top
            if not hasattr(self, funcname):
                continue
            else:
                self.do_pop()
                func = getattr(self, funcname)
                func()

if __name__ == '__main__':
    import sys
    try:
        import readline
    except ImportError:
        pass # not available on all platforms

    # If we see command-line arguments, interpret them as a stack state
    # and execute.  Otherwise go interactive.

    driver = PILDriver()
    if len(sys.argv[1:]) > 0:
        driver.execute(sys.argv[1:])
    else:
        print "PILDriver says hello."
        while 1:
            try:
                line = raw_input('pildriver> ');
            except EOFError:
                print "\nPILDriver says goodbye."
                break
            driver.execute(string.split(line))
            print driver.stack

# The following sets edit modes for GNU EMACS
# Local Variables:
# mode:python
# End:



from __future__ import unicode_literals
import json
from decimal import Decimal
import datetime
from django.http.response import Http404, HttpResponse
from django.test import TestCase, RequestFactory
from django.utils.translation import ugettext_lazy
import six
from advanced_reports.backoffice.api_utils import to_json, ViewRequestParameters
from advanced_reports.backoffice.base import BackOfficeBase


class SimpleBackOffice(BackOfficeBase):
    title = 'Simple Backoffice'

    def api_get_test(self, request):
        return request.view_params.get('param')

    def api_get_response(self, request):
        return HttpResponse('OK')


simple_backoffice = SimpleBackOffice()


class ApiTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def test_api(self):
        rf = RequestFactory()

        r = json.loads(simple_backoffice.api(rf.get('/api/')).content)
        self.assertEqual(r, None)

        with self.assertRaises(Http404):
            simple_backoffice.api(rf.get('/api/nonexisting/'), method='nonexisting')

        r = json.loads(simple_backoffice.api(rf.get('/api/test/?param=lol'), method='test').content)
        self.assertEqual(r['response_data'], 'lol')

        r = simple_backoffice.api(rf.get('/api/response/'), method='response').content
        self.assertEqual(r, 'OK')

    def test_to_json(self):
        self.assertEqual(to_json(123), '123')
        self.assertEqual(to_json('123'), '"123"')
        self.assertEqual(to_json({'a': '3', 'b': 5}), '{"a": "3", "b": 5}')
        self.assertEqual(to_json(Decimal('10.33')), '"10.33"')
        self.assertEqual(to_json(ugettext_lazy('ignore this also')), '"ignore this also"')
        self.assertEqual(to_json(datetime.datetime(2015, 1, 20)), '"2015-01-20T00:00:00"')
        self.assertEqual(to_json(datetime.date(2015, 1, 20)), '"2015-01-20"')

        @six.python_2_unicode_compatible
        class TestClass(object):
            def splitlines(self):
                return ('line 1', 'line 2')
            def __str__(self):
                return 'hello'

        class TestClass2(object):
            pass

        self.assertEqual(to_json(TestClass()), '"hello"')
        self.assertEqual(to_json(TestClass2()), 'null')

        with self.settings(DEBUG=True):
            self.assertEqual(to_json({'a': '3', 'b': 5}), '{\n  "a": "3", \n  "b": 5\n}')

    def test_viewrequestparameters(self):
        request = RequestFactory().get('/?a=1&a=2&a=3')
        params = ViewRequestParameters(request)
        self.assertEqual(params.get('nonexisting', 'fallback'), 'fallback')
        self.assertEqual(params.getlist('a'), ['1', '2', '3'])
        self.assertEqual(params.getlist('nonexisting'), [])
        self.assertEqual(params.getlist('nonexisting', ['default']), ['default'])
        self.assertEqual(repr(params),
                         "ViewRequestParameters(<QueryDict: {u'a': [u'1', u'2', u'3']}>, " +
                         "<QueryDict: {}>, {})")



from functools import wraps
from django.utils.translation import ugettext as _
from django.contrib.admin.forms import AdminAuthenticationForm
from django.contrib.auth.views import login
from django.contrib.auth import REDIRECT_FIELD_NAME

def staff_member_required(backoffice):
    def decorate(view_func):
        """
        Decorator for views that checks that the user is logged in and is a staff
        member, displaying the login page if necessary.
        """
        @wraps(view_func)
        def _checklogin(request, *args, **kwargs):
            if request.user.is_active and request.user.is_staff:
                # The user is valid. Continue to the admin page.
                return view_func(request, *args, **kwargs)

            assert hasattr(request, 'session'), "Advanced Reports Backoffice requires session middleware to be installed. Edit your MIDDLEWARE_CLASSES setting to insert 'django.contrib.sessions.middleware.SessionMiddleware'."
            defaults = {
                'template_name': backoffice.login_template,
                'authentication_form': AdminAuthenticationForm,
                'extra_context': {
                    'backoffice': backoffice,
                    REDIRECT_FIELD_NAME: request.get_full_path(),
                },
            }
            return login(request, **defaults)
        return _checklogin
    return decorate



from advanced_reports.backoffice.base import BackOfficeView

class SimpleView(BackOfficeView):
    def get_ip(self, request):
        return request.META['REMOTE_ADDR']

    def get_items(self, request):
        return {'a': 5, 'b': 3, 'c': 4}



from django.db import models
from django.conf import settings

# This is done for compatibility with other databases, mainly for testing.
if 'postgresql' in settings.DATABASES['default'].get('ENGINE', ''):
    from djorm_pgfulltext.fields import VectorField # pragma: no cover
    from djorm_pgfulltext.models import SearchManager # pragma: no cover
else:
    VectorField = lambda: u'' # pragma: no cover
    SearchManager = lambda *a, **kw: models.Manager() # pragma: no cover


class SearchIndex(models.Model):
    backoffice_instance = models.CharField(max_length=32)
    model_slug = models.CharField(max_length=32)
    model_id = models.PositiveIntegerField()
    to_index = models.TextField(blank=True)

    search_index = VectorField()

    objects = SearchManager(fields=('to_index',),
                            search_field='search_index',
                            auto_update_search_field=True)

    class Meta:
        verbose_name = u'search index entry'
        verbose_name_plural = u'search index entries'

    def __unicode__(self):
        return u'%s/%s/%d' % (self.backoffice_instance, self.model_slug, self.model_id)
